# This file is part of connectome-manipulator.
#
# SPDX-License-Identifier: Apache-2.0
# Copyright (c) 2024 Blue Brain Project/EPFL

import os
import re

import numpy as np
import pandas as pd
import pytest
from bluepysnap import Circuit
from mock import Mock, patch
from numpy.testing import assert_array_equal
from scipy.ndimage import gaussian_filter
from scipy.spatial import distance_matrix
from voxcell import VoxelData

from utils import TEST_DATA_DIR, setup_tempdir
import connectome_manipulator.model_building.conn_prob as test_module
from connectome_manipulator.model_building import model_types


def get_random_pos_matrix(n_row=None):
    """Get n_row positions between -50, 50"""
    n_row = n_row or np.random.randint(25, 50)
    return np.random.random((n_row, 3)) * 100 + 50


def test_extract():
    circuit = Circuit(os.path.join(TEST_DATA_DIR, "circuit_sonata.json"))

    functions = [
        "extract_1st_order",
        "extract_2nd_order",
        "extract_3rd_order",
        "extract_4th_order",
        "extract_5th_order",
    ]

    for i, f in enumerate(functions):
        with patch(f"connectome_manipulator.model_building.conn_prob.{f}") as patched:
            test_module.extract(circuit, i + 1, pos_map_file="")
            patched.assert_called()

    order = "fake"
    with pytest.raises(
        AssertionError, match=re.escape(f"Order-{order} data extraction not supported!")
    ):
        test_module.extract(circuit, order)


def test_build():
    functions = [
        "build_1st_order",
        "build_2nd_order",
        "build_3rd_order",
        "build_4th_order",
        "build_5th_order",
    ]

    for i, f in enumerate(functions):
        with patch(f"connectome_manipulator.model_building.conn_prob.{f}") as patched:
            test_module.build(i + 1)
            patched.assert_called()

    order = "fake"
    with pytest.raises(
        AssertionError, match=re.escape(f"Order-{order} model building not supported!")
    ):
        test_module.build(order)


def test_plot():
    functions = [
        "plot_1st_order",
        "plot_2nd_order",
        "plot_3rd_order",
        "plot_4th_order",
        "plot_5th_order",
    ]

    for i, f in enumerate(functions):
        with patch(f"connectome_manipulator.model_building.conn_prob.{f}") as patched:
            test_module.plot(i + 1)
            patched.assert_called()

    order = "fake"
    with pytest.raises(
        AssertionError, match=re.escape(f"Order-{order} data/model visualization not supported!")
    ):
        test_module.plot(order)


def test_load_pos_mapping_model():
    test_module.load_pos_mapping_model(None)

    with setup_tempdir(__name__) as tempdir:
        filepath = os.path.join(tempdir, "fake.json")

        with pytest.raises(
            AssertionError, match=re.escape("Position mapping model file not found!")
        ):
            test_module.load_pos_mapping_model(filepath)

        # Create dummy position mapping model
        pos_model = model_types.PosMapModel(pos_table=pd.DataFrame(np.random.rand(10, 3)))
        pos_model.save_model(
            os.path.split(filepath)[0], os.path.splitext(os.path.split(filepath)[1])[0]
        )

        test_module.load_pos_mapping_model(filepath)


def test_get_neuron_positions_by_id():
    functions = [
        lambda x: x + 0.1,
        lambda x: x * 2,
        lambda x: x**3,
    ]

    res = test_module.get_neuron_positions_by_id(functions, range(3))
    assert_array_equal(res, [0.1, 2, 8])

    res = test_module.get_neuron_positions_by_id(functions, [np.arange(3)] * 3)
    assert_array_equal(res, [[0.1, 1.1, 2.1], [0, 2, 4], [0, 1, 8]])

    res = test_module.get_neuron_positions_by_id(lambda x: x + 1, range(3))
    assert_array_equal(res, range(1, 4))


def test_get_neuron_positions():
    circuit = Circuit(os.path.join(TEST_DATA_DIR, "circuit_sonata.json"))
    nodes = [circuit.nodes["nodeA"]] * 2  # Src/tgt populations
    # Permute node IDs, so that src/tgt are different
    np.random.seed(0)
    src_ids = np.random.permutation(nodes[0].ids())
    tgt_ids = np.random.permutation(nodes[1].ids())

    # Case 1: No access function, no voxel map => Should correspond to node positions
    res = test_module.get_neuron_positions(nodes, [src_ids, tgt_ids], pos_acc=None, vox_map=None)
    assert_array_equal(nodes[0].positions(src_ids), res[0])
    assert_array_equal(nodes[1].positions(tgt_ids), res[1])

    # Case 2: Position access through PosMapModel (no voxel map) => Should correspond to mapped positions
    src_pos_table = pd.DataFrame(np.random.rand(len(src_ids), 3))
    tgt_pos_table = pd.DataFrame(np.random.rand(len(tgt_ids), 3))
    src_pos_model = model_types.PosMapModel(pos_table=src_pos_table)
    tgt_pos_model = model_types.PosMapModel(pos_table=tgt_pos_table)
    src_pos_acc = lambda gids: src_pos_model.apply(gids=gids)
    tgt_pos_acc = lambda gids: tgt_pos_model.apply(gids=gids)

    res = test_module.get_neuron_positions(
        nodes, [src_ids, tgt_ids], pos_acc=[src_pos_acc, tgt_pos_acc], vox_map=None
    )
    assert_array_equal(src_pos_table.loc[src_ids], res[0])
    assert_array_equal(tgt_pos_table.loc[tgt_ids], res[1])

    # Case 3: Position access through voxel map (no access function) => Should correspond to voxel positions
    src_vox_map = VoxelData.load_nrrd(os.path.join(TEST_DATA_DIR, "xy_map_lin.nrrd"))
    tgt_vox_map = VoxelData.load_nrrd(os.path.join(TEST_DATA_DIR, "xy_map_ones.nrrd"))

    res = test_module.get_neuron_positions(
        nodes, [src_ids, tgt_ids], pos_acc=None, vox_map=[src_vox_map, tgt_vox_map]
    )
    assert_array_equal(src_vox_map.lookup(nodes[0].positions(src_ids).to_numpy()), res[0])
    assert_array_equal(tgt_vox_map.lookup(nodes[1].positions(tgt_ids).to_numpy()), res[1])

    # Case 4: Both PosMapModel and voxel map provided => Must fail, not supported
    with pytest.raises(
        AssertionError,
        match=re.escape("Voxel map not supported when providing position access functions!"),
    ):
        res = test_module.get_neuron_positions(
            nodes,
            [src_ids, tgt_ids],
            pos_acc=[src_pos_acc, tgt_pos_acc],
            vox_map=[src_vox_map, tgt_vox_map],
        )


def test_get_pos_mapping_fcts():
    # Case 1: No mapping provided => Must return None's
    res = test_module.get_pos_mapping_fcts(pos_map_file=None)
    assert res[0] is None  # pos_acc
    assert res[1] is None  # vox_map

    # Case 2: Position mapping model .json file(s) => Must return pos_acc, but no vox_map
    with setup_tempdir(__name__) as tempdir:
        # Create dummy position mapping models
        pos_tables = []
        nids = 10
        np.random.seed(0)
        for i in range(2):
            _pos_tab = pd.DataFrame(np.random.rand(10, 3), index=i * nids + np.arange(nids))
            pos_model = model_types.PosMapModel(pos_table=_pos_tab)
            pos_model.save_model(tempdir, f"dummy_pos_map{i}")
            pos_tables.append(_pos_tab)

        # (a) Single src/tgt file
        res = test_module.get_pos_mapping_fcts(
            pos_map_file=os.path.join(tempdir, "dummy_pos_map0.json")
        )
        assert res[1] is None  # vox_map
        assert_array_equal(pos_tables[0].to_numpy(), res[0][0](pos_tables[0].index))  # src
        assert_array_equal(pos_tables[0].to_numpy(), res[0][1](pos_tables[0].index))  # tgt

        # (b) Separate src/tgt files
        res = test_module.get_pos_mapping_fcts(
            pos_map_file=[os.path.join(tempdir, f"dummy_pos_map{i}.json") for i in range(2)]
        )
        assert res[1] is None  # vox_map
        assert_array_equal(pos_tables[0].to_numpy(), res[0][0](pos_tables[0].index))  # src
        assert_array_equal(pos_tables[1].to_numpy(), res[0][1](pos_tables[1].index))  # tgt

    # Case 3: Voxel map .nrrd file(s) => Must return vox_map, but no pos_acc

    # (a) Single src/tgt file
    vox_file = os.path.join(TEST_DATA_DIR, "xy_map_lin.nrrd")
    vox_map = VoxelData.load_nrrd(vox_file)
    res = test_module.get_pos_mapping_fcts(pos_map_file=vox_file)
    assert res[0] is None  # pos_acc
    assert_array_equal(vox_map.raw, res[1][0].raw)  # src
    assert_array_equal(vox_map.bbox, res[1][0].bbox)
    assert_array_equal(vox_map.raw, res[1][1].raw)  # tgt
    assert_array_equal(vox_map.bbox, res[1][1].bbox)

    # (b) Separate src/tgt files
    src_vox_file = os.path.join(TEST_DATA_DIR, "xy_map_lin.nrrd")
    tgt_vox_file = os.path.join(TEST_DATA_DIR, "xy_map_ones.nrrd")
    src_vox_map = VoxelData.load_nrrd(src_vox_file)
    tgt_vox_map = VoxelData.load_nrrd(tgt_vox_file)
    res = test_module.get_pos_mapping_fcts(pos_map_file=[src_vox_file, tgt_vox_file])
    assert res[0] is None  # pos_acc
    assert_array_equal(src_vox_map.raw, res[1][0].raw)  # src
    assert_array_equal(src_vox_map.bbox, res[1][0].bbox)
    assert_array_equal(tgt_vox_map.raw, res[1][1].raw)  # tgt
    assert_array_equal(tgt_vox_map.bbox, res[1][1].bbox)

    # Case 4: Invalid file(s) => Must fail

    # (a) Single src/tgt file with wrong extension
    with pytest.raises(
        AssertionError, match=re.escape("Position mapping file error (must be .json or .nrrd)!")
    ):
        res = test_module.get_pos_mapping_fcts(pos_map_file="dummy.xyz")

    # (b) Separate src/tgt files with wrong extension
    with pytest.raises(
        AssertionError, match=re.escape("Position mapping file error (must be .json or .nrrd)!")
    ):
        res = test_module.get_pos_mapping_fcts(pos_map_file=["dummy1.xyz", "dummy2.xyz"])

    # (c) Separate src/tgt files with valid but different extensions
    with pytest.raises(
        AssertionError,
        match=re.escape("Same file type for source/target position mappings required!"),
    ):
        res = test_module.get_pos_mapping_fcts(pos_map_file=["dummy1.json", "dummy2.nrrd"])

    # (d) Single file provided as list
    with pytest.raises(
        AssertionError, match=re.escape("Two position mapping files (source/target) expected!")
    ):
        res = test_module.get_pos_mapping_fcts(pos_map_file=["dummy1.json"])

    # (e) More than two separte files provided
    with pytest.raises(
        AssertionError, match=re.escape("Two position mapping files (source/target) expected!")
    ):
        res = test_module.get_pos_mapping_fcts(
            pos_map_file=["dummy1.json", "dummy2.json", "dummy3.json"]
        )


def test_extract_dependant_p_conn():
    c = Circuit(os.path.join(TEST_DATA_DIR, "circuit_sonata.json"))
    edges = c.edges[c.edges.population_names[0]]
    src_ids = edges.source.ids()
    tgt_ids = edges.target.ids()

    # Case 1: No dependencies
    res_p_conn, res_count_conn, res_count_all = test_module.extract_dependent_p_conn(
        src_ids, tgt_ids, edges, [], []
    )
    assert res_count_all == len(src_ids) * len(
        tgt_ids
    ), "ERROR: Possible connection count mismatch!"
    assert res_count_conn == len(
        list(edges.iter_connections(src_ids, tgt_ids))
    ), "ERROR: Connection count mismatch!"
    assert res_p_conn == res_count_conn / res_count_all, "ERROR: Connection probability mismatch!"

    # Case 2: 1-D connection probabilities (2 bins only)
    dep_mat = np.array([np.arange(len(src_ids))]).T * np.arange(len(tgt_ids))
    dep_bins = np.linspace(np.min(dep_mat), np.max(dep_mat), 3)
    res_p_conn, res_count_conn, res_count_all = test_module.extract_dependent_p_conn(
        src_ids, tgt_ids, edges, [dep_mat], [dep_bins]
    )
    assert (
        len(res_p_conn) == len(res_count_conn) == len(res_count_all) == len(dep_bins) - 1
    ), "ERROR: Size mismatch!"
    adj_mat = np.array(
        [[len(list(edges.iter_connections(s, t))) > 0 for t in tgt_ids] for s in src_ids]
    )
    assert np.sum(res_count_all) == len(src_ids) * len(tgt_ids) and np.array_equal(
        res_count_all, np.histogram(dep_mat, dep_bins)[0]
    ), "ERROR: Possible connection count mismatch!"
    assert np.sum(res_count_conn) == np.sum(adj_mat) and np.array_equal(
        res_count_conn,
        [
            np.sum(np.logical_and(adj_mat, dep_mat < dep_bins[1])),
            np.sum(np.logical_and(adj_mat, dep_mat >= dep_bins[1])),
        ],
    ), "ERROR: Connection count mismatch!"
    assert np.array_equal(
        res_p_conn, res_count_conn / res_count_all
    ), "ERROR: Connection probability mismatch!"

    # Case 3: N-D connection probabilities (variable bins)
    N = 3
    dep_mats = [np.array([np.arange(len(src_ids))]).T * np.arange(len(tgt_ids))] * N
    dep_bins = [np.linspace(np.min(dep_mat), np.max(dep_mat), 3 * n + 5) for n in range(N)]
    res_p_conn, res_count_conn, res_count_all = test_module.extract_dependent_p_conn(
        src_ids, tgt_ids, edges, dep_mats, dep_bins
    )
    assert (
        res_p_conn.shape
        == res_count_conn.shape
        == res_count_all.shape
        == tuple([len(b) - 1 for b in dep_bins])
    ), "ERROR: Size mismatch!"
    assert np.sum(res_count_all) == len(src_ids) * len(
        tgt_ids
    ), "ERROR: Possible connection count mismatch!"
    assert np.sum(res_count_conn) == len(
        list(edges.iter_connections(src_ids, tgt_ids))
    ), "ERROR: Connection count mismatch!"
    assert np.array_equal(
        np.isnan(res_p_conn), np.isnan(res_count_conn / res_count_all)
    ) and np.array_equal(
        res_p_conn[np.isfinite(res_p_conn)],
        res_count_conn[res_count_all > 0] / res_count_all[res_count_all > 0],
    ), "ERROR: Connection probability mismatch!"


def test_get_value_ranges():
    np.random.seed(0)

    # Check special case: 0-dim
    ndim = 0
    rng = np.nan
    res = test_module.get_value_ranges(rng, ndim)
    assert len(res) == ndim

    # Check special case: 1-dim
    ndim = 1

    ## (a) Symmetric range
    rng = 100.0 * np.random.rand()
    res = test_module.get_value_ranges(rng, ndim)
    assert len(res) == 2
    assert res == [-rng, rng]

    ## (a) Arbitrary range
    rng = [-100.0 * np.random.rand(), 100.0 * np.random.rand()]
    res = test_module.get_value_ranges(rng, ndim)
    assert len(res) == 2
    assert res == [rng[0], rng[1]]

    # Check same range for all dims
    ndim = 10
    rng = 100.0 * np.random.rand()

    ## (a) Pos./neg. range
    res = test_module.get_value_ranges(rng, ndim)
    assert len(res) == ndim
    assert np.all([r == [-rng, rng] for r in res])

    ## (b) Pos. range only
    res = test_module.get_value_ranges(rng, ndim, True)
    assert len(res) == ndim
    assert np.all([r == [0, rng] for r in res])

    # Check different ranges for differnt dims
    rng = [100.0 * np.random.rand() for d in range(ndim)]

    ## (a) Pos./neg. range
    res = test_module.get_value_ranges(rng, ndim)
    assert len(res) == ndim
    assert np.all([res[i] == [-rng[i], rng[i]] for i in range(len(res))])

    ## (b) Pos. range only
    res = test_module.get_value_ranges(rng, ndim, True)
    assert len(res) == ndim
    assert np.all([res[i] == [0, rng[i]] for i in range(len(res))])

    ## (c) Mixed range
    pos_range = np.random.choice(2, ndim).astype(bool)
    res = test_module.get_value_ranges(rng, ndim, pos_range)
    assert len(res) == ndim
    assert np.all([res[i] == [0, rng[i]] for i in range(len(res)) if pos_range[i]])
    assert np.all([res[i] == [-rng[i], rng[i]] for i in range(len(res)) if not pos_range[i]])

    ## (d) Wrong numbers
    with pytest.raises(AssertionError, match=f"ERROR: max_range must have {ndim} elements!"):
        res = test_module.get_value_ranges(rng[:-1], ndim, pos_range)
    with pytest.raises(AssertionError, match=f"ERROR: pos_range must have {ndim} elements!"):
        res = test_module.get_value_ranges(rng, ndim, pos_range[:-1])

    # Check arbitrary ranges
    ndim = 10

    ## (a) Correct ranges (pos./neg.)
    rng = [[-100.0 * np.random.rand(), 100.0 * np.random.rand()] for i in range(ndim)]
    res = test_module.get_value_ranges(rng, ndim)
    assert len(res) == ndim
    assert np.all([res[i] == [rng[i][0], rng[i][1]] for i in range(len(res))])

    ## (b) Correct ranges (pos. only)
    rng = [[0.0, 100.0 * np.random.rand()] for i in range(ndim)]
    res = test_module.get_value_ranges(rng, ndim, True)
    assert len(res) == ndim
    assert np.all([res[i] == [rng[i][0], rng[i][1]] for i in range(len(res))])

    ## (c) Wrong pos./neg.
    rng = [[-100.0 * np.random.rand(), 100.0 * np.random.rand()] for i in range(ndim)]
    with pytest.raises(AssertionError, match=f"ERROR: Range of coord 0 must include 0!"):
        res = test_module.get_value_ranges(rng, ndim, True)

    ## (d) Wrong ranges
    rng = [[100.0 * np.random.rand(), -100.0 * np.random.rand()] for i in range(ndim)]
    with pytest.raises(AssertionError, match=f"ERROR: Range of coord 0 invalid!"):
        res = test_module.get_value_ranges(rng, ndim)


def test_extract_1st_order():
    circuit = Circuit(os.path.join(TEST_DATA_DIR, "circuit_sonata.json"))
    nodes = [circuit.nodes["nodeA"]] * 2  # Src/tgt populations
    edges = circuit.edges["nodeA__nodeA__chemical"]
    src_ids = nodes[0].ids()
    tgt_ids = nodes[1].ids()

    np.random.seed(0)
    for n in range(10):
        src_sel = np.random.choice(src_ids, np.random.choice(len(src_ids)) + 1, replace=False)
        tgt_sel = np.random.choice(tgt_ids, np.random.choice(len(tgt_ids)) + 1, replace=False)
        nsrc = len(src_sel)
        ntgt = len(tgt_sel)
        nconn = len(list(edges.iter_connections(source=src_sel, target=tgt_sel)))

        for min_nbins in [1, 10, 100]:
            res = test_module.extract_1st_order(
                nodes, edges, src_sel, tgt_sel, min_count_per_bin=min_nbins
            )
            if nsrc * ntgt >= min_nbins:
                assert np.isclose(res["p_conn"], nconn / (nsrc * ntgt))
            else:
                assert np.isnan(res["p_conn"])  # Not enought data points
            assert res["src_cell_count"] == nsrc
            assert res["tgt_cell_count"] == ntgt


def test_build_1st_order():
    np.random.seed(0)
    for p_conn in np.random.rand(10):
        model = test_module.build_1st_order(p_conn)
        assert np.isclose(model.p_conn, p_conn)


def test_extract_2nd_order():
    circuit = Circuit(os.path.join(TEST_DATA_DIR, "circuit_sonata.json"))
    nodes = [circuit.nodes["nodeA"]] * 2  # Src/tgt populations
    edges = circuit.edges["nodeA__nodeA__chemical"]
    src_ids = nodes[0].ids()
    tgt_ids = nodes[1].ids()

    bin_size_um = 100
    np.random.seed(0)
    for rep in range(10):
        src_sel = np.random.choice(src_ids, np.random.choice(len(src_ids)) + 1, replace=False)
        tgt_sel = np.random.choice(tgt_ids, np.random.choice(len(tgt_ids)) + 1, replace=False)
        nsrc = len(src_sel)
        ntgt = len(tgt_sel)

        src_pos = nodes[0].positions(src_sel)
        tgt_pos = nodes[0].positions(tgt_sel)
        dist = distance_matrix(src_pos, tgt_pos)  # Distance matrix
        dist[dist == 0.0] = np.nan
        nconn = np.array(
            [
                [len(list(edges.iter_connections(source=s, target=t))) for t in tgt_sel]
                for s in src_sel
            ]
        )  # Number of connection matrix

        num_bins = np.ceil(np.nanmax(dist) / bin_size_um).astype(int)  # Distance binning
        dist_bins = np.arange(0, num_bins + 1) * bin_size_um
        dist_bins[-1] += 1e-3  # So that max. value is always included in last bin
        conn_cnt = np.full(num_bins, -1)  # Conn. count
        all_cnt = np.full(num_bins, -1)  # All pair count
        for bidx in range(num_bins):
            dsel = np.logical_and(dist >= dist_bins[bidx], dist < dist_bins[bidx + 1])
            conn_cnt[bidx] = np.sum(nconn[dsel])
            all_cnt[bidx] = np.sum(dsel)
        p = conn_cnt / all_cnt  # Conn. prob.

        for min_nbins in [1, 3, 5]:
            res = test_module.extract_2nd_order(
                nodes, edges, src_sel, tgt_sel, bin_size_um=bin_size_um, min_count_per_bin=min_nbins
            )
            p_sel = p.copy()
            p_sel[all_cnt < min_nbins] = np.nan
            assert np.array_equal(res["p_conn_dist"], p_sel, equal_nan=True)
            assert np.array_equal(res["count_conn"], conn_cnt)
            assert np.array_equal(res["count_all"], all_cnt)
            assert np.array_equal(res["dist_bins"], dist_bins)
            assert res["src_cell_count"] == nsrc
            assert res["tgt_cell_count"] == ntgt


def test_build_2nd_order():
    dist_bins = np.arange(0, 1001, 10)
    d = np.array([np.mean(dist_bins[i : i + 2]) for i in range(len(dist_bins) - 1)])

    # Check simple exponential model building
    np.random.seed(0)
    for rep in range(10):
        exp_coefs = [1e-1 * np.random.rand(), 1e-2 * np.random.rand()]
        exp_model = lambda x, a, b: a * np.exp(-b * np.array(x))
        exp_data = exp_model(d, *exp_coefs)

        model = test_module.build_2nd_order(
            exp_data, dist_bins, np.zeros_like(exp_data), model_specs={"type": "SimpleExponential"}
        )
        model_coefs = [model.get_param_dict()[k] for k in ["scale", "exponent"]]
        assert np.allclose(exp_coefs, model_coefs)

    # Check complex exponential model building [EXPERIMENTAL: Just one working test example, since model fitting not so robust]
    exp_coefs = [1e-1, 1e-4, 2.0, 1e-1, 1e-4]
    exp_model = lambda x, a, b, c, d, e: a * np.exp(-b * np.array(x) ** c) + d * np.exp(
        -e * np.array(x)
    )
    exp_data = exp_model(d, *exp_coefs)

    model = test_module.build_2nd_order(
        exp_data, dist_bins, np.zeros_like(exp_data), model_specs={"type": "ComplexExponential"}
    )
    model_coefs = [
        model.get_param_dict()[k]
        for k in ["prox_scale", "prox_exp", "prox_exp_pow", "dist_scale", "dist_exp"]
    ]
    assert np.allclose(exp_coefs, model_coefs)


def test_extract_3rd_order():
    circuit = Circuit(os.path.join(TEST_DATA_DIR, "circuit_sonata.json"))
    nodes = [circuit.nodes["nodeA"]] * 2  # Src/tgt populations
    edges = circuit.edges["nodeA__nodeA__chemical"]
    src_ids = nodes[0].ids()
    tgt_ids = nodes[1].ids()

    bin_size_um = 100
    np.random.seed(0)
    for rep in range(10):
        src_sel = np.random.choice(src_ids, np.random.choice(len(src_ids)) + 1, replace=False)
        tgt_sel = np.random.choice(tgt_ids, np.random.choice(len(tgt_ids)) + 1, replace=False)
        nsrc = len(src_sel)
        ntgt = len(tgt_sel)

        src_pos = nodes[0].positions(src_sel)
        tgt_pos = nodes[0].positions(tgt_sel)
        dist = distance_matrix(src_pos, tgt_pos)  # Distance matrix
        dist[dist == 0.0] = np.nan
        bip = np.array(
            [[np.sign(tgt_pos["z"].loc[t] - src_pos["z"].loc[s]) for t in tgt_sel] for s in src_sel]
        )
        nconn = np.array(
            [
                [len(list(edges.iter_connections(source=s, target=t))) for t in tgt_sel]
                for s in src_sel
            ]
        )  # Number of connection matrix

        num_bins = np.ceil(np.nanmax(dist) / bin_size_um).astype(int)  # Distance binning
        dist_bins = np.arange(0, num_bins + 1) * bin_size_um
        dist_bins[-1] += 1e-3  # So that max. value is always included in last bin
        conn_cnt = np.full((num_bins, 2), -1)  # Conn. count
        all_cnt = np.full((num_bins, 2), -1)  # All pair count
        for bidx in range(num_bins):
            for bipidx, bipval in enumerate([-1, 1]):
                dsel = np.logical_and(
                    np.logical_and(dist >= dist_bins[bidx], dist < dist_bins[bidx + 1]),
                    bip == bipval,
                )
                conn_cnt[bidx, bipidx] = np.sum(nconn[dsel])
                all_cnt[bidx, bipidx] = np.sum(dsel)
        p = conn_cnt / all_cnt  # Conn. prob.

        for min_nbins in [1, 3, 5]:
            res = test_module.extract_3rd_order(
                nodes, edges, src_sel, tgt_sel, bin_size_um=bin_size_um, min_count_per_bin=min_nbins
            )
            p_sel = p.copy()
            p_sel[all_cnt < min_nbins] = np.nan
            assert np.array_equal(res["p_conn_dist_bip"], p_sel, equal_nan=True)
            assert np.array_equal(res["count_conn"], conn_cnt)
            assert np.array_equal(res["count_all"], all_cnt)
            assert np.array_equal(res["dist_bins"], dist_bins)
            assert res["src_cell_count"] == nsrc
            assert res["tgt_cell_count"] == ntgt


def test_build_3rd_order():
    dist_bins = np.arange(0, 1001, 10)
    d = np.array([np.mean(dist_bins[i : i + 2]) for i in range(len(dist_bins) - 1)])

    # Check simple exponential model building
    np.random.seed(0)
    for rep in range(10):
        exp_coefs = [
            1e-1 * np.random.rand(),
            1e-2 * np.random.rand(),
            1e-1 * np.random.rand(),
            1e-2 * np.random.rand(),
        ]  # 'scale_N', 'exponent_N', 'scale_P', 'exponent_P'
        exp_model = lambda x, a, b: a * np.exp(-b * np.array(x))
        exp_data = np.array([exp_model(d, *exp_coefs[:2]), exp_model(d, *exp_coefs[2:])]).T

        model = test_module.build_3rd_order(
            exp_data,
            dist_bins,
            np.zeros_like(exp_data),
            bip_coord_data=0,
            model_specs={"type": "SimpleExponential"},
        )
        model_coefs = [
            model.get_param_dict()[k] for k in ["scale_N", "exponent_N", "scale_P", "exponent_P"]
        ]
        assert np.allclose(exp_coefs, model_coefs)

    # Check complex exponential model building [EXPERIMENTAL: Just one working test example, since model fitting not so robust]
    exp_coefs = [
        1e-1,
        1e-4,
        2.0,
        1e-1,
        1e-4,
        2e-1,
        2e-4,
        1.75,
        2e-1,
        2e-4,
    ]  # 'prox_scale_N', 'prox_exp_N', 'prox_exp_pow_N', 'dist_scale_N', 'dist_exp_N', 'prox_scale_P', 'prox_exp_P', 'prox_exp_pow_P', 'dist_scale_P', 'dist_exp_P'
    exp_model = lambda x, a, b, c, d, e: a * np.exp(-b * np.array(x) ** c) + d * np.exp(
        -e * np.array(x)
    )
    exp_data = np.array([exp_model(d, *exp_coefs[:5]), exp_model(d, *exp_coefs[5:])]).T

    model = test_module.build_3rd_order(
        exp_data,
        dist_bins,
        np.zeros_like(exp_data),
        bip_coord_data=0,
        model_specs={"type": "ComplexExponential"},
    )
    model_coefs = [
        model.get_param_dict()[k]
        for k in [
            "prox_scale_N",
            "prox_exp_N",
            "prox_exp_pow_N",
            "dist_scale_N",
            "dist_exp_N",
            "prox_scale_P",
            "prox_exp_P",
            "prox_exp_pow_P",
            "dist_scale_P",
            "dist_exp_P",
        ]
    ]
    assert np.allclose(exp_coefs, model_coefs)


def test_extract_4th_order():
    circuit = Circuit(os.path.join(TEST_DATA_DIR, "circuit_sonata.json"))
    nodes = [circuit.nodes["nodeA"]] * 2  # Src/tgt populations
    edges = circuit.edges["nodeA__nodeA__chemical"]
    src_ids = nodes[0].ids()
    tgt_ids = nodes[1].ids()

    bin_size_um = 200
    np.random.seed(0)
    for rep in range(5):
        src_sel = np.random.choice(src_ids, np.random.choice(len(src_ids)) + 1, replace=False)
        tgt_sel = np.random.choice(tgt_ids, np.random.choice(len(tgt_ids)) + 1, replace=False)
        nsrc = len(src_sel)
        ntgt = len(tgt_sel)

        src_pos = nodes[0].positions(src_sel)
        tgt_pos = nodes[1].positions(tgt_sel)
        offmat = np.array(
            [
                [
                    [tgt_pos[coord].loc[t] - src_pos[coord].loc[s] for coord in ["x", "y", "z"]]
                    for t in tgt_sel
                ]
                for s in src_sel
            ]
        )
        nconn = np.array(
            [
                [len(list(edges.iter_connections(source=s, target=t))) for t in tgt_sel]
                for s in src_sel
            ]
        )  # Number of connection matrix

        ranges = [
            [np.nanmin(offmat[:, :, coord]), np.nanmax(offmat[:, :, coord])]
            for coord in range(offmat.shape[2])
        ]
        num_bins = [
            np.ceil((ranges[coord][1] - ranges[coord][0]) / bin_size_um).astype(int)
            for coord in range(len(ranges))
        ]
        off_bins = [
            np.arange(0, num_bins[coord] + 1) * bin_size_um + ranges[coord][0]
            for coord in range(len(ranges))
        ]

        conn_cnt = np.full(num_bins, -1)  # Conn. count
        all_cnt = np.full(num_bins, -1)  # All pair count
        for xidx in range(num_bins[0]):
            for yidx in range(num_bins[1]):
                for zidx in range(num_bins[2]):
                    xsel = np.logical_and(
                        offmat[:, :, 0] >= off_bins[0][xidx],
                        offmat[:, :, 0] < off_bins[0][xidx + 1],
                    )
                    ysel = np.logical_and(
                        offmat[:, :, 1] >= off_bins[1][yidx],
                        offmat[:, :, 1] < off_bins[1][yidx + 1],
                    )
                    zsel = np.logical_and(
                        offmat[:, :, 2] >= off_bins[2][zidx],
                        offmat[:, :, 2] < off_bins[2][zidx + 1],
                    )
                    sel = np.logical_and(np.logical_and(xsel, ysel), zsel)
                    conn_cnt[xidx, yidx, zidx] = np.sum(nconn[sel])
                    all_cnt[xidx, yidx, zidx] = np.sum(sel)
        p = conn_cnt / all_cnt  # Conn. prob.

        for min_nbins in [1, 3, 5]:
            res = test_module.extract_4th_order(
                nodes, edges, src_sel, tgt_sel, bin_size_um=bin_size_um, min_count_per_bin=min_nbins
            )
            p_sel = p.copy()
            p_sel[all_cnt < min_nbins] = np.nan
            assert np.array_equal(res["p_conn_offset"], p_sel, equal_nan=True)
            assert np.array_equal(res["count_conn"], conn_cnt)
            assert np.array_equal(res["count_all"], all_cnt)
            assert np.all(
                [
                    np.allclose(res[k], off_bins[i])
                    for i, k in enumerate(["dx_bins", "dy_bins", "dz_bins"])
                ]
            )
            assert res["src_cell_count"] == nsrc
            assert res["tgt_cell_count"] == ntgt


def test_build_4th_order():
    np.random.seed(0)
    bin_sizes = [100, 200, 400]
    dx_bins = np.arange(0, 1001, bin_sizes[0])
    dy_bins = np.arange(0, 1001, bin_sizes[1])
    dz_bins = np.arange(0, 1001, bin_sizes[2])
    dx = np.array([np.mean(dx_bins[i : i + 2]) for i in range(len(dx_bins) - 1)])
    dy = np.array([np.mean(dy_bins[i : i + 2]) for i in range(len(dy_bins) - 1)])
    dz = np.array([np.mean(dz_bins[i : i + 2]) for i in range(len(dz_bins) - 1)])
    p = 1e-2 * np.random.rand(len(dx_bins) - 1, len(dy_bins) - 1, len(dz_bins) - 1)

    # Check linear interpolation model building
    model = test_module.build_4th_order(
        p, dx_bins, dy_bins, dz_bins, np.zeros_like(p), model_specs={"type": "LinearInterpolation"}
    )
    p_model = np.array(
        [
            [[model.get_conn_prob(dx=_dx, dy=_dy, dz=_dz)[0] for _dz in dz] for _dy in dy]
            for _dx in dx
        ]
    )
    assert np.allclose(p, p_model)

    # Check linear interpolation model building with Gaussian filtering
    smoothing_sigma_um = 100.0
    sigmas = [smoothing_sigma_um / b for b in bin_sizes]
    p_filt = gaussian_filter(p, sigmas, mode="constant")
    model = test_module.build_4th_order(
        p,
        dx_bins,
        dy_bins,
        dz_bins,
        np.zeros_like(p),
        model_specs={"type": "LinearInterpolation"},
        smoothing_sigma_um=smoothing_sigma_um,
    )
    p_model = np.array(
        [
            [[model.get_conn_prob(dx=_dx, dy=_dy, dz=_dz)[0] for _dz in dz] for _dy in dy]
            for _dx in dx
        ]
    )
    assert np.allclose(p_filt, p_model)

    # Check random forest regressor model building [Not yet implemented]
    with pytest.raises(AssertionError, match='ERROR: Model type "RandomForestRegressor" unknown!'):
        model = test_module.build_4th_order(
            p,
            dx_bins,
            dy_bins,
            dz_bins,
            np.zeros_like(p),
            model_specs={"type": "RandomForestRegressor"},
        )


def test_extract_4th_order_reduced():
    circuit = Circuit(os.path.join(TEST_DATA_DIR, "circuit_sonata.json"))
    nodes = [circuit.nodes["nodeA"]] * 2  # Src/tgt populations
    edges = circuit.edges["nodeA__nodeA__chemical"]
    src_ids = nodes[0].ids()
    tgt_ids = nodes[1].ids()

    bin_size_um = 200
    np.random.seed(0)
    for rep in range(5):
        src_sel = np.random.choice(src_ids, np.random.choice(len(src_ids)) + 1, replace=False)
        tgt_sel = np.random.choice(tgt_ids, np.random.choice(len(tgt_ids)) + 1, replace=False)
        nsrc = len(src_sel)
        ntgt = len(tgt_sel)

        src_pos = nodes[0].positions(src_sel)
        tgt_pos = nodes[0].positions(tgt_sel)
        offmat = np.array(
            [
                [
                    [tgt_pos[coord].loc[t] - src_pos[coord].loc[s] for coord in ["x", "y", "z"]]
                    for t in tgt_sel
                ]
                for s in src_sel
            ]
        )
        offmat = np.stack(
            [np.sqrt(offmat[:, :, 0] ** 2 + offmat[:, :, 1] ** 2), offmat[:, :, 2]], axis=2
        )
        nconn = np.array(
            [
                [len(list(edges.iter_connections(source=s, target=t))) for t in tgt_sel]
                for s in src_sel
            ]
        )  # Number of connection matrix

        ranges = [
            [0.0, np.nanmax(offmat[:, :, 0])],
            [np.nanmin(offmat[:, :, 1]), np.nanmax(offmat[:, :, 1])],
        ]
        num_bins = [
            np.ceil((ranges[coord][1] - ranges[coord][0]) / bin_size_um).astype(int)
            for coord in range(len(ranges))
        ]
        off_bins = [
            np.arange(0, num_bins[coord] + 1) * bin_size_um + ranges[coord][0]
            for coord in range(len(ranges))
        ]

        conn_cnt = np.full(num_bins, -1)  # Conn. count
        all_cnt = np.full(num_bins, -1)  # All pair count
        for ridx in range(num_bins[0]):
            for zidx in range(num_bins[1]):
                rsel = np.logical_and(
                    offmat[:, :, 0] >= off_bins[0][ridx], offmat[:, :, 0] < off_bins[0][ridx + 1]
                )
                zsel = np.logical_and(
                    offmat[:, :, 1] >= off_bins[1][zidx], offmat[:, :, 1] < off_bins[1][zidx + 1]
                )
                sel = np.logical_and(rsel, zsel)
                conn_cnt[ridx, zidx] = np.sum(nconn[sel])
                all_cnt[ridx, zidx] = np.sum(sel)
        p = conn_cnt / all_cnt  # Conn. prob.

        for min_nbins in [1, 3, 5]:
            res = test_module.extract_4th_order_reduced(
                nodes, edges, src_sel, tgt_sel, bin_size_um=bin_size_um, min_count_per_bin=min_nbins
            )
            p_sel = p.copy()
            p_sel[all_cnt < min_nbins] = np.nan
            assert np.array_equal(res["p_conn_offset"], p_sel, equal_nan=True)
            assert np.array_equal(res["count_conn"], conn_cnt)
            assert np.array_equal(res["count_all"], all_cnt)
            assert np.all(
                [np.allclose(res[k], off_bins[i]) for i, k in enumerate(["dr_bins", "dz_bins"])]
            )
            assert res["src_cell_count"] == nsrc
            assert res["tgt_cell_count"] == ntgt


def test_build_4th_order_reduced():
    np.random.seed(0)
    bin_sizes = [100, 200]
    dr_bins = np.arange(0, 1001, bin_sizes[0])
    dz_bins = np.arange(0, 1001, bin_sizes[1])
    dr = np.array([np.mean(dr_bins[i : i + 2]) for i in range(len(dr_bins) - 1)])
    dz = np.array([np.mean(dz_bins[i : i + 2]) for i in range(len(dz_bins) - 1)])
    p = 1e-2 * np.random.rand(len(dr_bins) - 1, len(dz_bins) - 1)

    # Check linear interpolation model building
    model = test_module.build_4th_order_reduced(
        p,
        dr_bins,
        dz_bins,
        np.zeros_like(p),
        axial_coord_data=0,
        model_specs={"type": "LinearInterpolation"},
    )
    p_model = np.array([[model.get_conn_prob(dr=_dr, dz=_dz)[0] for _dz in dz] for _dr in dr])
    assert np.allclose(p, p_model)

    # Check linear interpolation model building with Gaussian filtering
    smoothing_sigma_um = 100.0
    sigmas = [smoothing_sigma_um / b for b in bin_sizes]
    p_reflect = np.vstack(
        [p[::-1, :], p]
    )  # Mirror along radial axis at dr==0, to avoid edge effect
    p_reflect = gaussian_filter(p_reflect, sigmas, mode="constant")
    p_filt = p_reflect[p.shape[0] :, :]  # Cut original part of the data
    model = test_module.build_4th_order_reduced(
        p,
        dr_bins,
        dz_bins,
        np.zeros_like(p),
        axial_coord_data=0,
        model_specs={"type": "LinearInterpolation"},
        smoothing_sigma_um=smoothing_sigma_um,
    )
    p_model = np.array([[model.get_conn_prob(dr=_dr, dz=_dz)[0] for _dz in dz] for _dr in dr])
    assert np.allclose(p_filt, p_model)

    # Check random forest regressor model building [Not yet implemented]
    with pytest.raises(AssertionError, match='ERROR: Model type "RandomForestRegressor" unknown!'):
        model = test_module.build_4th_order_reduced(
            p,
            dr_bins,
            dz_bins,
            np.zeros_like(p),
            axial_coord_data=0,
            model_specs={"type": "RandomForestRegressor"},
        )


def test_extract_5th_order():
    circuit = Circuit(os.path.join(TEST_DATA_DIR, "circuit_sonata.json"))
    nodes = [circuit.nodes["nodeA"]] * 2  # Src/tgt populations
    edges = circuit.edges["nodeA__nodeA__chemical"]
    src_ids = nodes[0].ids()
    tgt_ids = nodes[1].ids()

    off_bin_size_um = 500
    pos_bin_size_um = 1000
    np.random.seed(0)
    for rep in range(5):
        src_sel = np.random.choice(src_ids, np.random.choice(len(src_ids)) + 1, replace=False)
        tgt_sel = np.random.choice(tgt_ids, np.random.choice(len(tgt_ids)) + 1, replace=False)
        nsrc = len(src_sel)
        ntgt = len(tgt_sel)

        src_pos = nodes[0].positions(src_sel)
        tgt_pos = nodes[0].positions(tgt_sel)
        posmat = np.array(
            [np.repeat(src_pos[[coord]].to_numpy(), ntgt, axis=1).T for coord in ["x", "y", "z"]]
        ).T
        posmat2 = np.array(
            [np.repeat(tgt_pos[[coord]].to_numpy(), ntgt, axis=1).T for coord in ["x", "y", "z"]]
        ).T
        offmat = np.array(
            [
                [
                    [tgt_pos[coord].loc[t] - src_pos[coord].loc[s] for coord in ["x", "y", "z"]]
                    for t in tgt_sel
                ]
                for s in src_sel
            ]
        )
        nconn = np.array(
            [
                [len(list(edges.iter_connections(source=s, target=t))) for t in tgt_sel]
                for s in src_sel
            ]
        )  # Number of connection matrix

        pos_ranges = [
            [
                np.minimum(np.nanmin(posmat[:, :, coord]), np.nanmin(posmat2[:, :, coord])),
                np.maximum(np.nanmax(posmat[:, :, coord]), np.nanmax(posmat2[:, :, coord])),
            ]
            for coord in range(posmat.shape[2])
        ]  # Set range based on pre- AND post-neurons
        off_ranges = [
            [np.nanmin(offmat[:, :, coord]), np.nanmax(offmat[:, :, coord])]
            for coord in range(offmat.shape[2])
        ]
        pos_num_bins = [
            np.ceil((pos_ranges[coord][1] - pos_ranges[coord][0]) / pos_bin_size_um).astype(int)
            for coord in range(len(pos_ranges))
        ]
        off_num_bins = [
            np.ceil((off_ranges[coord][1] - off_ranges[coord][0]) / off_bin_size_um).astype(int)
            for coord in range(len(off_ranges))
        ]
        pos_bins = [
            np.arange(0, pos_num_bins[coord] + 1) * pos_bin_size_um + pos_ranges[coord][0]
            for coord in range(len(pos_ranges))
        ]
        off_bins = [
            np.arange(0, off_num_bins[coord] + 1) * off_bin_size_um + off_ranges[coord][0]
            for coord in range(len(off_ranges))
        ]

        conn_cnt = np.full(pos_num_bins + off_num_bins, -1)  # Conn. count
        all_cnt = np.full(pos_num_bins + off_num_bins, -1)  # All pair count
        for xidx in range(pos_num_bins[0]):
            for yidx in range(pos_num_bins[1]):
                for zidx in range(pos_num_bins[2]):
                    for dxidx in range(off_num_bins[0]):
                        for dyidx in range(off_num_bins[1]):
                            for dzidx in range(off_num_bins[2]):
                                xsel = np.logical_and(
                                    posmat[:, :, 0] >= pos_bins[0][xidx],
                                    posmat[:, :, 0] < pos_bins[0][xidx + 1],
                                )
                                ysel = np.logical_and(
                                    posmat[:, :, 1] >= pos_bins[1][yidx],
                                    posmat[:, :, 1] < pos_bins[1][yidx + 1],
                                )
                                zsel = np.logical_and(
                                    posmat[:, :, 2] >= pos_bins[2][zidx],
                                    posmat[:, :, 2] < pos_bins[2][zidx + 1],
                                )
                                dxsel = np.logical_and(
                                    offmat[:, :, 0] >= off_bins[0][dxidx],
                                    offmat[:, :, 0] < off_bins[0][dxidx + 1],
                                )
                                dysel = np.logical_and(
                                    offmat[:, :, 1] >= off_bins[1][dyidx],
                                    offmat[:, :, 1] < off_bins[1][dyidx + 1],
                                )
                                dzsel = np.logical_and(
                                    offmat[:, :, 2] >= off_bins[2][dzidx],
                                    offmat[:, :, 2] < off_bins[2][dzidx + 1],
                                )
                                sel = np.logical_and(
                                    np.logical_and(
                                        np.logical_and(xsel, dxsel), np.logical_and(ysel, dysel)
                                    ),
                                    np.logical_and(zsel, dzsel),
                                )
                                conn_cnt[xidx, yidx, zidx, dxidx, dyidx, dzidx] = np.sum(nconn[sel])
                                all_cnt[xidx, yidx, zidx, dxidx, dyidx, dzidx] = np.sum(sel)
        p = conn_cnt / all_cnt  # Conn. prob.

        for min_nbins in [1, 3, 5]:
            res = test_module.extract_5th_order(
                nodes,
                edges,
                src_sel,
                tgt_sel,
                position_bin_size_um=pos_bin_size_um,
                offset_bin_size_um=off_bin_size_um,
                min_count_per_bin=min_nbins,
            )
            p_sel = p.copy()
            p_sel[all_cnt < min_nbins] = np.nan
            assert np.array_equal(res["p_conn_position"], p_sel, equal_nan=True)
            assert np.array_equal(res["count_conn"], conn_cnt)
            assert np.array_equal(res["count_all"], all_cnt)
            assert np.all(
                [
                    np.array_equal(res[k], pos_bins[i])
                    for i, k in enumerate(["x_bins", "y_bins", "z_bins"])
                ]
            )
            assert np.all(
                [
                    np.allclose(res[k], off_bins[i])
                    for i, k in enumerate(["dx_bins", "dy_bins", "dz_bins"])
                ]
            )
            assert res["src_cell_count"] == nsrc
            assert res["tgt_cell_count"] == ntgt


def test_build_5th_order():
    np.random.seed(0)
    bin_sizes = [500, 500, 500, 200, 300, 400]
    x_bins = np.arange(0, 1001, bin_sizes[0])
    y_bins = np.arange(0, 1001, bin_sizes[1])
    z_bins = np.arange(0, 2001, bin_sizes[2])
    dx_bins = np.arange(0, 1001, bin_sizes[3])
    dy_bins = np.arange(0, 1001, bin_sizes[4])
    dz_bins = np.arange(0, 1001, bin_sizes[5])
    x = np.array([np.mean(x_bins[i : i + 2]) for i in range(len(x_bins) - 1)])
    y = np.array([np.mean(y_bins[i : i + 2]) for i in range(len(y_bins) - 1)])
    z = np.array([np.mean(z_bins[i : i + 2]) for i in range(len(z_bins) - 1)])
    dx = np.array([np.mean(dx_bins[i : i + 2]) for i in range(len(dx_bins) - 1)])
    dy = np.array([np.mean(dy_bins[i : i + 2]) for i in range(len(dy_bins) - 1)])
    dz = np.array([np.mean(dz_bins[i : i + 2]) for i in range(len(dz_bins) - 1)])
    p = 1e-2 * np.random.rand(
        len(x_bins) - 1,
        len(y_bins) - 1,
        len(z_bins) - 1,
        len(dx_bins) - 1,
        len(dy_bins) - 1,
        len(dz_bins) - 1,
    )

    # Check linear interpolation model building
    model = test_module.build_5th_order(
        p,
        x_bins,
        y_bins,
        z_bins,
        dx_bins,
        dy_bins,
        dz_bins,
        np.zeros_like(p),
        model_specs={"type": "LinearInterpolation"},
    )
    p_model = np.array(
        [
            [
                [
                    [
                        [
                            [
                                model.get_conn_prob(x=_x, y=_y, z=_z, dx=_dx, dy=_dy, dz=_dz)[0]
                                for _dz in dz
                            ]
                            for _dy in dy
                        ]
                        for _dx in dx
                    ]
                    for _z in z
                ]
                for _y in y
            ]
            for _x in x
        ]
    )
    assert np.allclose(p, p_model)

    # Check linear interpolation model building with Gaussian filtering
    smoothing_sigma_um = 100.0
    sigmas = [smoothing_sigma_um / b for b in bin_sizes]
    p_filt = gaussian_filter(p, sigmas, mode="constant")
    model = test_module.build_5th_order(
        p,
        x_bins,
        y_bins,
        z_bins,
        dx_bins,
        dy_bins,
        dz_bins,
        np.zeros_like(p),
        model_specs={"type": "LinearInterpolation"},
        smoothing_sigma_um=smoothing_sigma_um,
    )
    p_model = np.array(
        [
            [
                [
                    [
                        [
                            [
                                model.get_conn_prob(x=_x, y=_y, z=_z, dx=_dx, dy=_dy, dz=_dz)[0]
                                for _dz in dz
                            ]
                            for _dy in dy
                        ]
                        for _dx in dx
                    ]
                    for _z in z
                ]
                for _y in y
            ]
            for _x in x
        ]
    )
    assert np.allclose(p_filt, p_model)

    # Check random forest regressor model building [Not yet implemented]
    with pytest.raises(AssertionError, match='ERROR: Model type "RandomForestRegressor" unknown!'):
        model = test_module.build_5th_order(
            p,
            x_bins,
            y_bins,
            z_bins,
            dx_bins,
            dy_bins,
            dz_bins,
            np.zeros_like(p),
            model_specs={"type": "RandomForestRegressor"},
        )


def test_extract_5th_order_reduced():
    circuit = Circuit(os.path.join(TEST_DATA_DIR, "circuit_sonata.json"))
    nodes = [circuit.nodes["nodeA"]] * 2  # Src/tgt populations
    edges = circuit.edges["nodeA__nodeA__chemical"]
    src_ids = nodes[0].ids()
    tgt_ids = nodes[1].ids()

    off_bin_size_um = 250
    pos_bin_size_um = 1000
    np.random.seed(0)
    for rep in range(5):
        src_sel = np.random.choice(src_ids, np.random.choice(len(src_ids)) + 1, replace=False)
        tgt_sel = np.random.choice(tgt_ids, np.random.choice(len(tgt_ids)) + 1, replace=False)
        nsrc = len(src_sel)
        ntgt = len(tgt_sel)

        src_pos = nodes[0].positions(src_sel)
        tgt_pos = nodes[0].positions(tgt_sel)
        posmat = np.repeat(src_pos[["z"]].to_numpy(), ntgt, axis=1)
        posmat2 = np.repeat(tgt_pos[["z"]].to_numpy(), ntgt, axis=1)
        offmat = np.array(
            [
                [
                    [tgt_pos[coord].loc[t] - src_pos[coord].loc[s] for coord in ["x", "y", "z"]]
                    for t in tgt_sel
                ]
                for s in src_sel
            ]
        )
        offmat = np.stack(
            [np.sqrt(offmat[:, :, 0] ** 2 + offmat[:, :, 1] ** 2), offmat[:, :, 2]], axis=2
        )
        nconn = np.array(
            [
                [len(list(edges.iter_connections(source=s, target=t))) for t in tgt_sel]
                for s in src_sel
            ]
        )  # Number of connection matrix

        pos_ranges = [
            np.minimum(np.nanmin(posmat), np.nanmin(posmat2)),
            np.maximum(np.nanmax(posmat), np.nanmax(posmat2)),
        ]  # Set range based on pre- AND post-neurons
        off_ranges = [
            [0.0, np.nanmax(offmat[:, :, 0])],
            [np.nanmin(offmat[:, :, 1]), np.nanmax(offmat[:, :, 1])],
        ]
        pos_num_bins = np.ceil((pos_ranges[1] - pos_ranges[0]) / pos_bin_size_um).astype(int)
        off_num_bins = [
            np.ceil((off_ranges[coord][1] - off_ranges[coord][0]) / off_bin_size_um).astype(int)
            for coord in range(len(off_ranges))
        ]
        pos_bins = np.arange(0, pos_num_bins + 1) * pos_bin_size_um + pos_ranges[0]
        off_bins = [
            np.arange(0, off_num_bins[coord] + 1) * off_bin_size_um + off_ranges[coord][0]
            for coord in range(len(off_ranges))
        ]

        conn_cnt = np.full([pos_num_bins] + off_num_bins, -1)  # Conn. count
        all_cnt = np.full([pos_num_bins] + off_num_bins, -1)  # All pair count
        for zidx in range(pos_num_bins):
            for dridx in range(off_num_bins[0]):
                for dzidx in range(off_num_bins[1]):
                    zsel = np.logical_and(posmat >= pos_bins[zidx], posmat < pos_bins[zidx + 1])
                    drsel = np.logical_and(
                        offmat[:, :, 0] >= off_bins[0][dridx],
                        offmat[:, :, 0] < off_bins[0][dridx + 1],
                    )
                    dzsel = np.logical_and(
                        offmat[:, :, 1] >= off_bins[1][dzidx],
                        offmat[:, :, 1] < off_bins[1][dzidx + 1],
                    )
                    sel = np.logical_and(zsel, np.logical_and(drsel, dzsel))
                    conn_cnt[zidx, dridx, dzidx] = np.sum(nconn[sel])
                    all_cnt[zidx, dridx, dzidx] = np.sum(sel)
        p = conn_cnt / all_cnt  # Conn. prob.

        for min_nbins in [1, 3, 5]:
            res = test_module.extract_5th_order_reduced(
                nodes,
                edges,
                src_sel,
                tgt_sel,
                position_bin_size_um=pos_bin_size_um,
                offset_bin_size_um=off_bin_size_um,
                min_count_per_bin=min_nbins,
            )
            p_sel = p.copy()
            p_sel[all_cnt < min_nbins] = np.nan
            assert np.array_equal(res["p_conn_position"], p_sel, equal_nan=True)
            assert np.array_equal(res["count_conn"], conn_cnt)
            assert np.array_equal(res["count_all"], all_cnt)
            assert np.array_equal(res["z_bins"], pos_bins)
            assert np.all(
                [np.allclose(res[k], off_bins[i]) for i, k in enumerate(["dr_bins", "dz_bins"])]
            )
            assert res["src_cell_count"] == nsrc
            assert res["tgt_cell_count"] == ntgt


def test_build_5th_order_reduced():
    np.random.seed(0)
    bin_sizes = [500, 100, 200]
    z_bins = np.arange(0, 2001, bin_sizes[0])
    dr_bins = np.arange(0, 1001, bin_sizes[1])
    dz_bins = np.arange(0, 1001, bin_sizes[2])
    z = np.array([np.mean(z_bins[i : i + 2]) for i in range(len(z_bins) - 1)])
    dr = np.array([np.mean(dr_bins[i : i + 2]) for i in range(len(dr_bins) - 1)])
    dz = np.array([np.mean(dz_bins[i : i + 2]) for i in range(len(dz_bins) - 1)])
    p = 1e-2 * np.random.rand(len(z_bins) - 1, len(dr_bins) - 1, len(dz_bins) - 1)

    # Check linear interpolation model building
    model = test_module.build_5th_order_reduced(
        p,
        z_bins,
        dr_bins,
        dz_bins,
        np.zeros_like(p),
        axial_coord_data=0,
        model_specs={"type": "LinearInterpolation"},
    )
    p_model = np.array(
        [[[model.get_conn_prob(z=_z, dr=_dr, dz=_dz)[0] for _dz in dz] for _dr in dr] for _z in z]
    )
    assert np.allclose(p, p_model)

    # Check linear interpolation model building with Gaussian filtering
    smoothing_sigma_um = 100.0
    sigmas = [smoothing_sigma_um / b for b in bin_sizes]
    p_reflect = np.concatenate(
        [p[:, ::-1, :], p], axis=1
    )  # Mirror along radial axis at dr==0, to avoid edge effect
    p_reflect = gaussian_filter(p_reflect, sigmas, mode="constant")
    p_filt = p_reflect[:, p.shape[1] :, :]  # Cut original part of the data
    model = test_module.build_5th_order_reduced(
        p,
        z_bins,
        dr_bins,
        dz_bins,
        np.zeros_like(p),
        axial_coord_data=0,
        model_specs={"type": "LinearInterpolation"},
        smoothing_sigma_um=smoothing_sigma_um,
    )
    p_model = np.array(
        [[[model.get_conn_prob(z=_z, dr=_dr, dz=_dz)[0] for _dz in dz] for _dr in dr] for _z in z]
    )
    assert np.allclose(p_filt, p_model)

    # Check random forest regressor model building [Not yet implemented]
    with pytest.raises(AssertionError, match='ERROR: Model type "RandomForestRegressor" unknown!'):
        model = test_module.build_5th_order_reduced(
            p,
            z_bins,
            dr_bins,
            dz_bins,
            np.zeros_like(p),
            axial_coord_data=0,
            model_specs={"type": "RandomForestRegressor"},
        )



# This file is part of connectome-manipulator.
#
# SPDX-License-Identifier: Apache-2.0
# Copyright (c) 2024 Blue Brain Project/EPFL

"""Main module for connectome manipulations:

- Loads a SONATA connectome using SNAP
- Applies manipulation(s) to the connectome, as specified by the manipulation config dict
- Writes back the manipulated connectome to a SONATA edges file, together with a new circuit config
"""

import copy
import os
from dataclasses import dataclass
from datetime import datetime
from pathlib import Path
import logging

import numpy as np
import pandas as pd
import bluepysnap
from bluepysnap.circuit import Circuit
import libsonata

from .. import log, utils, profiler
from ..access_functions import get_edges_population, get_nodes_population
from ..processing import BatchInfo, get_node_splits
from . import executors
from .tracker import JobTracker
from .converters import (
    create_parquet_metadata,
    parquet_to_sonata,
    EdgeWriter,
    _SYNAPSE_PROPERTIES,
    _PROPERTY_TYPES,
    _ID_COLUMN_MAP,
)
from .manipulation import Manipulation

logger = logging.getLogger(__name__)
_ID_COLUMN_INV_MAP = {_v: _k for _k, _v in _ID_COLUMN_MAP.items()}


@dataclass
class Options:
    """CLI Options which have defaults"""

    output_path: Path
    config_path: Path
    logging_path: Path
    do_profiling: bool = False
    do_resume: bool = False
    keep_parquet: bool = False
    convert_to_sonata: bool = False
    overwrite_edges: bool = False
    splits: int = 0
    target_payload: int = 20_000_000_000
    parallel: bool = False


@dataclass
class JobsCommonInfo:
    """Information shared across all jobs"""

    nodes: [bluepysnap.nodes.Nodes]
    edges: [bluepysnap.edges.Edges]


@dataclass
class JobInfo:
    """An abstraction for a chunk of work, defined by its index and total"""

    split_index: int
    split_total: int
    batches: [BatchInfo]
    out_parquet_file: str

    @property
    def flat_size(self):
        """The total number of node ids this job is supposed to process"""
        return sum(len(b.node_ids) for b in self.batches)

    def flatten(self):
        """Retuns a flat list with all node ids this job is supposed to process"""
        return np.concatenate([b.node_ids for b in self.batches]).astype(np.int64)


def load_circuit(
    sonata_config, existing_edge_popul_name=None, src_node_popul_name=None, tgt_node_popul_name=None
):
    """Load given edges population of a SONATA circuit using SNAP."""
    # Load circuit
    log.info(f"Loading circuit from {sonata_config}")
    c = Circuit(sonata_config)

    if (
        hasattr(c, "edges")
        and hasattr(c.edges, "population_names")
        and len(c.edges.population_names) > 0
    ):
        # Select edge population
        edges, popul_name = get_edges_population(
            c, existing_edge_popul_name, return_popul_name=True
        )
        edges_file = c.config["networks"]["edges"][0]["edges_file"]

        # Select corresponding source/target nodes populations
        src_nodes = edges.source
        tgt_nodes = edges.target

        if src_node_popul_name is not None:
            log.log_assert(
                src_node_popul_name == src_nodes.name,
                "Source nodes population name not consistent with selected edges population!",
            )
        if tgt_node_popul_name is not None:
            log.log_assert(
                tgt_node_popul_name == tgt_nodes.name,
                "Target nodes population name not consistent with selected edges population!",
            )

        if src_nodes is not tgt_nodes:
            log.log_assert(
                len(np.intersect1d(src_nodes.ids(), tgt_nodes.ids())) == 0,
                "Node IDs must be unique across different node populations!",
            )
    else:  # Circuit w/o edges
        edges = None
        edges_file = None
        popul_name = None

        src_nodes = get_nodes_population(c, src_node_popul_name)
        tgt_nodes = get_nodes_population(c, tgt_node_popul_name)

    nodes = (src_nodes, tgt_nodes)

    src_nodes_file = c.to_libsonata.node_population_properties(src_nodes.name).elements_path
    tgt_nodes_file = c.to_libsonata.node_population_properties(tgt_nodes.name).elements_path
    nodes_files = [src_nodes_file, tgt_nodes_file]

    if edges is None:
        log.debug(
            f'No edges population defined between nodes "{src_nodes.name}" and "{tgt_nodes.name}"'
        )
    else:
        log.debug(
            f'Using edges population "{edges.name}" between nodes "{src_nodes.name}" and "{tgt_nodes.name}"'
        )

    return c.config, nodes, nodes_files, edges, edges_file, popul_name


def apply_manipulation(edges_table, nodes, job: JobInfo, manip: dict):
    """Apply manipulation to connectome (edges_table) as specified in the manip_config."""
    log.info(f'Applying manipulation "{manip["name"]}" for split {job.split_index}')
    log.info(f'Results will be written to "{job.out_parquet_file}"')

    if edges_table is None:
        if "syn_props_init" in manip:
            # Initialize edges table with specified properties + default ones (excluding "edge_type_id" which is for internal use)
            syn_props_dict = {
                _ID_COLUMN_INV_MAP.get(_prop, _prop): _PROPERTY_TYPES[_prop]
                for _prop in _SYNAPSE_PROPERTIES
                if _prop != "edge_type_id"
            }
            syn_props_dict.update(manip["syn_props_init"])
            edges_table = pd.DataFrame(
                {_prop: pd.Series(dtype=_type) for _prop, _type in syn_props_dict.items()}
            )
    else:
        log.log_assert(
            "syn_props_init" not in manip,
            '"syn_props_init" not supported with existing connectome!',
        )

    with EdgeWriter(job.out_parquet_file, existing_edges=edges_table) as writer:
        for fun, cfg in enumerate(manip["fcts"]):
            source = cfg.pop("source")
            models = cfg.pop("model_config")

            log.info(f">>Function {fun + 1} of {len(manip['fcts'])}: source={source}")

            if filename := cfg.pop("model_pathways", None):
                pathways = pd.read_parquet(filename)
                if "sel_src" in cfg or "sel_dest" in cfg:
                    raise KeyError(
                        "Cannot specify pathway configuration and sel_src/sel_dest at the same time."
                    )
            else:
                pathways = None

            m = Manipulation.get(source)(nodes, writer, job.split_index, job.split_total)

            for batch in job.batches:
                for n, (node_ids, sel_src, sel_dest, pathway_specs) in enumerate(
                    batch.process_pathways(pathways)
                ):
                    kwargs = cfg | models
                    if sel_src:
                        kwargs["sel_src"] = sel_src
                    if sel_dest:
                        kwargs["sel_dest"] = sel_dest

                    log.info(
                        f">>Step {n + 1}: source={source} sel_src={kwargs.get('sel_src')} sel_dest={kwargs.get('sel_dest')}"
                    )

                    m.apply(
                        libsonata.Selection(node_ids).flatten(),
                        pathway_specs=pathway_specs,
                        **kwargs,
                    )
        return len(writer)


def create_new_file_from_template(new_file, template_file, replacements_dict, skip_comments=True):
    """Create new text file from template with replacements."""
    log.info(f"Creating file {new_file}")
    with open(template_file, "r") as file:
        content = file.read()

    content_lines = []
    for line in content.splitlines():
        if (
            skip_comments and len(line.strip()) > 0 and line.strip()[0] == "#"
        ):  # Skip replacement in commented lines
            content_lines.append(line)
        else:  # Apply replacements
            for src, dest in replacements_dict.items():
                line = line.replace(src, dest)
            content_lines.append(line)
    content = "\n".join(content_lines)

    with open(new_file, "w") as file:
        file.write(content)


def create_sonata_config(
    existing_config, out_config_path, out_edges_path, population_name, src_popul_name=None
):
    """Create new SONATA config (.JSON) from original, incl. modifications."""
    log.info(f"Creating SONATA config {out_config_path}")

    config = copy.deepcopy(existing_config)

    existing_edge_list = config["networks"].get("edges")
    if existing_edge_list is None:
        existing_edge_list = []
    if src_popul_name is not None:
        existing_edge_list = [
            _e for _e in existing_edge_list if src_popul_name not in _e["populations"]
        ]  # Remove source edges population, if existing
    new_edges = {
        "edges_file": str(out_edges_path),
        "populations": {population_name: {"type": "chemical"}},
    }
    config["networks"]["edges"] = [
        new_edges
    ] + existing_edge_list  # Add new (manipulated) population

    config = utils.reduce_config_paths(config, config_dir=Path(out_config_path).parent)

    utils.write_json(data=config, filepath=out_config_path)


def create_workflow_config(circuit_path, blue_config, manip_name, output_path, template_file):
    """Create bbp-workflow config for circuit registration (from template)."""
    if manip_name is None:
        manip_name = ""

    if len(manip_name) > 0:
        workflow_file = (
            os.path.split(os.path.splitext(template_file)[0])[1]
            + f"_{manip_name}"
            + os.path.splitext(template_file)[1]
        )
        circuit_name = "_".join(circuit_path.split("/")[-4:] + [manip_name])
        circuit_descr = f"{manip_name} applied to {circuit_path}"
        circuit_type = "Circuit manipulated by connectome_manipulator"
    else:
        workflow_file = os.path.split(template_file)[1]
        circuit_name = "_".join(circuit_path.split("/")[-4:])
        circuit_descr = f"No manipulation applied to {circuit_path}"
        circuit_type = "Circuit w/o manipulation"

    workflow_path = os.path.join(output_path, "workflows")
    if not os.path.exists(workflow_path):
        os.makedirs(workflow_path)

    config_replacements = {
        "$CIRCUIT_NAME": circuit_name,
        "$CIRCUIT_DESCRIPTION": circuit_descr,
        "$CIRCUIT_TYPE": circuit_type,
        "$CIRCUIT_CONFIG": blue_config,
        "$DATE": datetime.today().strftime("%Y-%m-%d %H:%M:%S") + " [generated from template]",
        "$FILE_NAME": workflow_file,
    }
    create_new_file_from_template(
        os.path.join(workflow_path, workflow_file),
        template_file,
        config_replacements,
        skip_comments=False,
    )


def manip_wrapper(jobs_common: JobsCommonInfo, job: JobInfo, options: Options):
    """Wrapper function (remote) that can be optionally executed by a Dask worker"""
    if options.parallel:
        import socket

        path = os.path.join(options.logging_path, socket.getfqdn())
        log.create_log_file(path, f"connectome_manipulation.task-{job.split_index}")

    log.info("Processing job %s (common: %s)", job, jobs_common)
    profiler.ProfilerManager.set_enabled(options.do_profiling)
    config = utils.load_json(options.config_path)
    np.random.seed(config.get("seed", 123456) * (job.split_index + 1))

    # Apply connectome wiring
    edges_table = _get_afferent_edges_table(job.flatten(), jobs_common.edges)

    for idx in range(len(config["manip"]["fcts"])):
        if filename := config["manip"]["fcts"][idx].get("model_pathways"):
            config["manip"]["fcts"][idx]["model_pathways"] = options.config_path.parent / filename

    N_syn_in = len(edges_table) if edges_table is not None else 0

    with profiler.profileit(name="processing"):
        N_syn_out = apply_manipulation(
            edges_table,
            jobs_common.nodes,
            job,
            config["manip"],
        )

    return N_syn_in, N_syn_out, profiler.ProfilerManager


@profiler.profileit(name="connectome_manipulation_main")
def main(options, log_file, executor_args=()):
    """Build local connectome."""
    config = utils.load_json(options.config_path)
    sonata_config_file = config["circuit_config"]
    edges_popul_name = config.get("edges_popul_name")
    src_node_popul_name = config.get("src_node_popul_name")
    tgt_node_popul_name = config.get("tgt_node_popul_name")

    if "circuit_path" in config:
        circuit_path = Path(config["circuit_path"])
        sonata_config_file = circuit_path / sonata_config_file
    else:
        circuit_path = Path(os.path.split(config["circuit_config"])[0])
    log.log_assert(options.output_path != circuit_path, "Input directory == Output directory")

    # Initialize the profiler
    profiler.ProfilerManager.set_csv_file(csv_file=log_file + ".csv")

    sonata_config, nodes, _, edges, _, src_popul_name = load_circuit(
        sonata_config_file, edges_popul_name, src_node_popul_name, tgt_node_popul_name
    )

    # Define target node splits
    node_ids_split = get_node_splits(config, options, nodes)

    edges_file = options.output_path / "edges.h5"
    log.log_assert(
        options.overwrite_edges or not edges_file.exists(),
        f'Edges file "{edges_file}" already exists! Enable "overwrite_edges" flag to overwrite!',
    )

    # Prepare params to the executor. Slurm executor will add the prefix itself
    executor_params = dict(x.split("=", 1) for x in executor_args)
    tracker = JobTracker(options.output_path, len(node_ids_split))

    with tracker.follow_jobs() as result_hook:
        with executors.in_context(options, executor_params, result_hook=result_hook) as executor:
            log.info("Start job submission")
            jobs_common = JobsCommonInfo(nodes, edges)

            for i_split, parquet_file in tracker.prepare_parquet_dir(options.do_resume):
                job = JobInfo(i_split, len(node_ids_split), node_ids_split[i_split], parquet_file)
                log.info(
                    f"Split {job.split_index + 1}/{len(node_ids_split)}: Wiring connectome targeting {job.flat_size} neurons"
                )
                executor.submit(
                    manip_wrapper,
                    (jobs_common, job, options),
                    {"out_parquet_file": job.out_parquet_file},
                )

    if options.convert_to_sonata:
        # Follow SONATA convention for edge population naming
        edge_population_name = f"{nodes[0].name}__{nodes[1].name}__chemical"
        # [IMPORTANT: .parquet to SONATA converter requires same column data types in all files!! Otherwise, value over-/underflows may occur due to wrong interpretation of numbers!!]
        parquet_to_sonata(
            tracker.parquet_dir,
            edges_file,
            nodes,
            tracker.parquet_done_file,
            population_name=edge_population_name,
            keep_parquet=options.keep_parquet,
        )

        # Create new SONATA config (.JSON) from original config file
        out_config_path = Path(options.output_path, "circuit_config.json")
        create_sonata_config(
            existing_config=sonata_config,
            out_config_path=out_config_path,
            out_edges_path=edges_file,
            population_name=edge_population_name,
            src_popul_name=src_popul_name,
        )
    else:
        if not options.keep_parquet:
            log.warning(
                "--keep-parquet and --convert-to-sonata are set to false. I will keep the parquet files anyway"
            )
        create_parquet_metadata(tracker.parquet_dir, nodes)


def _get_afferent_edges_table(node_ids, edges):
    if edges is None:
        return None
    return edges.afferent_edges(node_ids, properties=sorted(edges.property_names)).astype(
        edges.property_dtypes
    )


def _write_blue_config(manip_config, output_path, edges_fn_manip, edges_file_manip):
    # Create new symlinks and circuit config
    if not manip_config.get("blue_config_to_update") is None:
        blue_config = os.path.join(
            manip_config["circuit_path"], manip_config["blue_config_to_update"]
        )
        log.log_assert(
            os.path.exists(blue_config),
            f'Blue config "{manip_config["blue_config_to_update"]}" does not exist!',
        )
        with open(blue_config, "r") as file:  # Read blue config
            config = file.read()
        nrn_path = (
            list(
                filter(
                    lambda x: x.find("nrnPath") >= 0 and not x.strip()[0] == "#",
                    config.splitlines(),
                )
            )[0]
            .replace("nrnPath", "")
            .strip()
        )  # Extract path to edges file from BlueConfig
        circ_path_entry = list(
            filter(
                lambda x: x.find("CircuitPath") >= 0 and not x.strip()[0] == "#",
                config.splitlines(),
            )
        )[
            0
        ].strip()  # Extract circuit path entry from BlueConfig
        log.log_assert(
            os.path.abspath(nrn_path).find(os.path.abspath(manip_config["circuit_path"])) == 0,
            "nrnPath not within circuit path!",
        )
        nrn_path_manip = os.path.join(
            output_path, os.path.relpath(nrn_path, manip_config["circuit_path"])
        )  # Re-based path
        if not os.path.exists(os.path.split(nrn_path_manip)[0]):
            os.makedirs(os.path.split(nrn_path_manip)[0])

        # Symbolic link for edges.sonata
        symlink_src = os.path.relpath(edges_file_manip, os.path.split(nrn_path_manip)[0])
        symlink_dst = os.path.join(
            os.path.split(nrn_path_manip)[0], os.path.splitext(edges_fn_manip)[0] + ".sonata"
        )
        if os.path.isfile(symlink_dst) or os.path.islink(symlink_dst):
            os.remove(symlink_dst)  # Remove if already exists
        os.symlink(symlink_src, symlink_dst)
        log.info(f"Creating symbolic link ...{symlink_dst} -> {symlink_src}")

        # Create BlueConfig for manipulated circuit
        blue_config_manip = os.path.join(
            output_path,
            os.path.splitext(manip_config["blue_config_to_update"])[0]
            + f'_{manip_config["manip"]["name"]}'
            + os.path.splitext(manip_config["blue_config_to_update"])[1],
        )
        config_replacement = {
            nrn_path: symlink_dst,  # BETTER (???): nrn_path_manip
            circ_path_entry: f"CircuitPath {output_path}",
        }
        create_new_file_from_template(blue_config_manip, blue_config, config_replacement)

        # Symbolic link for start.target (if not existing)
        symlink_src = os.path.join(manip_config["circuit_path"], "start.target")
        symlink_dst = os.path.join(output_path, "start.target")
        if os.path.isfile(symlink_src) and not os.path.isfile(symlink_dst):
            os.symlink(symlink_src, symlink_dst)
            log.info(f"Creating symbolic link ...{symlink_dst} -> {symlink_src}")

        # Symbolic link for CellLibraryFile (if not existing)
        cell_lib_fn = (
            list(filter(lambda x: x.find("CellLibraryFile") >= 0, config.splitlines()))[0]
            .replace("CellLibraryFile", "")
            .strip()
        )  # Extract cell library file from BlueConfig
        if len(os.path.split(cell_lib_fn)[0]) == 0:  # Filename only, no path
            symlink_src = os.path.join(manip_config["circuit_path"], cell_lib_fn)
            symlink_dst = os.path.join(output_path, cell_lib_fn)
            if os.path.isfile(symlink_src) and not os.path.isfile(symlink_dst):
                os.symlink(symlink_src, symlink_dst)
                log.info(f"Creating symbolic link ...{symlink_dst} -> {symlink_src}")

        # Create bbp-workflow config from template to register manipulated circuit
        if manip_config.get("workflow_template") is not None:
            if os.path.exists(manip_config["workflow_template"]):
                create_workflow_config(
                    manip_config["circuit_path"],
                    blue_config_manip,
                    manip_config["manip"]["name"],
                    output_path,
                    manip_config["workflow_template"],
                )
            else:
                log.error(
                    f'Unable to create workflow config! Workflow template file "{manip_config["workflow_template"]}" not found!'
                )



# This file is part of connectome-manipulator.
#
# SPDX-License-Identifier: Apache-2.0
# Copyright (c) 2024 Blue Brain Project/EPFL

import connectome_manipulator.connectome_manipulation.manipulation.syn_removal as test_module
import numpy as np
import os
import pytest

from bluepysnap import Circuit
from numpy.testing import assert_allclose, assert_approx_equal, assert_array_equal

from utils import TEST_DATA_DIR
from connectome_manipulator.connectome_manipulation.manipulation import Manipulation
from connectome_manipulator.connectome_manipulation.converters import EdgeWriter
from connectome_manipulator import log


@pytest.fixture
def manipulation():
    m = Manipulation.get("syn_removal")
    return m


def test_apply(manipulation):
    log.setup_logging()  # To have data logging in a defined state

    circuit = Circuit(os.path.join(TEST_DATA_DIR, "circuit_sonata.json"))
    edges = circuit.edges[circuit.edges.population_names[0]]
    nodes = [edges.source, edges.target]
    src_ids = nodes[0].ids()
    tgt_ids = nodes[1].ids()
    edges_table = edges.afferent_edges(tgt_ids, properties=edges.property_names)

    # Check that corrent number of synapses is removed
    total_syns = len(edges_table)
    for pct in range(0, 101, 25):
        writer = EdgeWriter(None, existing_edges=edges_table.copy())
        manipulation(nodes, writer).apply(tgt_ids, amount_pct=pct)
        res = writer.to_pandas()
        assert len(res) == int(total_syns * (100 - pct) / 100)

    # Check that only given ids are considered
    src_id = 0
    writer = EdgeWriter(None, existing_edges=edges_table.copy())
    manipulation(nodes, writer).apply(tgt_ids, sel_src={*src_ids} - {src_id}, amount_pct=100)
    res = writer.to_pandas()
    assert np.all(res["@source_node"] == src_id)

    tgt_id = 9
    writer = EdgeWriter(None, existing_edges=edges_table.copy())
    manipulation(nodes, writer).apply(tgt_ids, sel_dest={*tgt_ids} - {tgt_id}, amount_pct=100)
    res = writer.to_pandas()
    assert np.all(res["@target_node"] == tgt_id)

    # Check when both sel_src and sel_dest are given
    writer = EdgeWriter(None, existing_edges=edges_table.copy())
    manipulation(nodes, writer).apply(
        tgt_ids, sel_src={*src_ids} - {src_id}, sel_dest={*tgt_ids} - {tgt_id}, amount_pct=100
    )
    res = writer.to_pandas()
    assert np.all(np.logical_or(res["@source_node"] == src_id, res["@target_node"] == tgt_id))

    # Check that one synapse per connection is left when keep-cons=True
    total_conns = len(np.unique(edges_table[["@source_node", "@target_node"]], axis=0))
    writer = EdgeWriter(None, existing_edges=edges_table.copy())
    manipulation(nodes, writer).apply(tgt_ids, amount_pct=100, keep_conns=True, rescale_gsyn=True)
    res = writer.to_pandas()
    res_conn = np.unique(res[["@source_node", "@target_node"]], axis=0)
    assert len(res_conn) == total_conns
    assert len(res) == total_conns

    # Check that conductance per connection is maintained with rescale_gsyn=True
    assert_allclose(
        res.groupby(["@source_node", "@target_node"])["conductance"].sum(),
        edges_table.groupby(["@source_node", "@target_node"])["conductance"].sum(),
    )


def test_get_gsyn_sum_per_conn():
    circuit = Circuit(os.path.join(TEST_DATA_DIR, "circuit_sonata.json"))
    edges = circuit.edges[circuit.edges.population_names[0]]
    nodes = [edges.source, edges.target]
    src_ids = nodes[0].ids()
    tgt_ids = nodes[1].ids()
    edges_table = edges.afferent_edges(tgt_ids, properties=edges.property_names)

    # Test full edges table
    gsyn_table = test_module.SynapseRemoval._get_gsyn_sum_per_conn(edges_table, src_ids, tgt_ids)
    edge_groups = edges_table.groupby(["@source_node", "@target_node"])
    for sgid in src_ids:
        for tgid in tgt_ids:
            if (sgid, tgid) in edge_groups.groups:
                assert_approx_equal(
                    gsyn_table[sgid, tgid], edge_groups.get_group((sgid, tgid)).conductance.sum()
                )

    # Test subset
    src_sel = [3, 9]
    tgt_sel = [2, 4, 7]
    gsyn_table = test_module.SynapseRemoval._get_gsyn_sum_per_conn(edges_table, src_sel, tgt_sel)
    mask = np.logical_and(
        np.in1d(edges_table["@source_node"], src_sel),
        np.in1d(edges_table["@target_node"], tgt_sel),
    )
    assert np.sum(mask) > 0 and np.sum(mask) < len(mask), "Bad choice for testing!"
    edge_groups = edges_table.loc[mask].groupby(["@source_node", "@target_node"])
    for sgid in src_sel:
        for tgid in tgt_sel:
            if (sgid, tgid) in edge_groups.groups:
                assert_approx_equal(
                    gsyn_table[sgid, tgid], edge_groups.get_group((sgid, tgid)).conductance.sum()
                )


def test_rescale_gsyn_per_conn():
    circuit = Circuit(os.path.join(TEST_DATA_DIR, "circuit_sonata.json"))
    edges = circuit.edges[circuit.edges.population_names[0]]
    nodes = [edges.source, edges.target]
    src_ids = nodes[0].ids()
    tgt_ids = nodes[1].ids()
    edges_table_orig = edges.afferent_edges(tgt_ids, properties=edges.property_names)

    # Basically test rescaling the conductance of every connection by a factor of 2
    ratio = 2
    edges_table = edges_table_orig.copy()
    gsyn_table = test_module.SynapseRemoval._get_gsyn_sum_per_conn(edges_table, src_ids, tgt_ids)
    gsyn_table_manip = np.array(gsyn_table) * ratio

    test_module.SynapseRemoval._rescale_gsyn_per_conn(
        edges_table, src_ids, tgt_ids, gsyn_table, gsyn_table_manip
    )
    assert_array_equal(edges_table.conductance, edges_table_orig.conductance / ratio)

    # Now the same but only for a subset
    src_sel = [3, 9]
    tgt_sel = [2, 4, 7]
    edges_table = edges_table_orig.copy()
    gsyn_table = test_module.SynapseRemoval._get_gsyn_sum_per_conn(edges_table, src_sel, tgt_sel)
    gsyn_table_manip = np.array(gsyn_table) * ratio

    test_module.SynapseRemoval._rescale_gsyn_per_conn(
        edges_table, src_sel, tgt_sel, gsyn_table, gsyn_table_manip
    )

    # Check that only the connections defined by node_ids are changed
    mask = np.logical_and(
        np.in1d(edges_table["@source_node"], src_sel),
        np.in1d(edges_table["@target_node"], tgt_sel),
    )
    assert np.sum(mask) > 0 and np.sum(mask) < len(mask), "Bad choice for testing!"
    assert_array_equal(edges_table.conductance[mask], edges_table_orig.conductance[mask] / ratio)
    assert_array_equal(edges_table.conductance[~mask], edges_table_orig.conductance[~mask])



# This file is part of connectome-manipulator.
#
# SPDX-License-Identifier: Apache-2.0
# Copyright (c) 2024 Blue Brain Project/EPFL

"""Module for building (deterministic) connection probability models based on adjacency matrices"""

import os.path

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from scipy.sparse import csc_matrix

from connectome_manipulator import log
from connectome_manipulator.connectome_comparison import adjacency
from connectome_manipulator.model_building import model_types


def extract(circuit, sel_src=None, sel_dest=None, edges_popul_name=None, CV_dict=None, **_):
    """Extracts adjacency matrix between selected src/dest neurons.

    Args:
        circuit (bluepysnap.Circuit): Input circuit
        sel_src (str/list-like/dict): Source (pre-synaptic) neuron selection
        sel_dest (str/list-like/dict): Target (post-synaptic) neuron selection
        edges_popul_name (str): Name of SONATA egdes population to extract data from
        CV_dict (dict): Cross-validation dictionary - Not supported

    Returns:
        dict: Dictionary containing the extracted adjacency matrix and source/target node ids
    """
    log.info(f"Running adjacency data extraction (sel_src={sel_src}, sel_dest={sel_dest})...")
    log.log_assert(CV_dict is None, "ERROR: Cross-validation not supported!")

    adj_dict = adjacency.compute(circuit, sel_src, sel_dest, edges_popul_name)

    return {
        "adj_mat": adj_dict["adj"]["data"].tocsc(),
        "src_node_ids": adj_dict["common"]["src_gids"],
        "tgt_node_ids": adj_dict["common"]["tgt_gids"],
    }


def build(adj_mat, src_node_ids, tgt_node_ids, inverted=False, **_):
    """Builds a (deterministic) connection probability model of type ``ConnProbAdjModel`` from an adjacency matrix (i.e., returning probabilities 0.0 or 1.0 only).

    Args:
        adj_mat (scipy.sparse.csc_matrix): Sparse adjacency matrix with boolean entries (i.e., True...connection, False...no connection)
        src_node_ids (list-like): List of source (pre-synaptic) neuron IDs
        tgt_node_ids (list-like): List of target (post-synaptic) neuron IDs
        inverted (bool): Flag for interpreting the boolean matrix entries in an inverted way (i.e., True...no connection, False...connection)

    Returns:
        connectome_manipulator.model_building.model_types.ConnProbAdjModel: Resulting adjacency model
    """
    log.info(f"Running {'inverted ' if inverted else ''}adjacency model building...")

    log.log_assert(
        adj_mat.dtype == "bool", "ERROR: Adjacency matrix with boolean data type required!"
    )
    log.log_assert(
        adj_mat.format.lower() == "csc", "ERROR: Adjacency matrix in CSC format required!"
    )
    log.log_assert(adj_mat.shape[0] == len(src_node_ids), "ERROR: Source nodes mismatch!")
    log.log_assert(adj_mat.shape[1] == len(tgt_node_ids), "ERROR: Target nodes mismatch!")

    # Prepare data frames
    src_nodes_table = pd.DataFrame(src_node_ids, columns=["src_node_ids"])
    tgt_nodes_table = pd.DataFrame(tgt_node_ids, columns=["tgt_node_ids"])
    rows, cols = adj_mat.nonzero()
    adj_table = pd.DataFrame({"row_ind": rows, "col_ind": cols})

    # Create model
    model = model_types.ConnProbAdjModel(
        src_nodes_table=src_nodes_table,
        tgt_nodes_table=tgt_nodes_table,
        adj_table=adj_table,
        inverted=inverted,
    )
    log.debug("Model description:\n%s", model)

    return model


def plot(out_dir, model, **_):
    """Visualizes the adjacency model.

    Args:
        out_dir (str): Path to output directory where the results figures will be stored
        model (connectome_manipulator.model_building.model_types.ConnProbAdjModel): Adjacency model, as returned by :func:`build`
    """
    log.info("Running adjacency model visualization...")

    # Apply model, i.e., getting connection probabilities between all pairs of neurons
    src_node_ids = model.get_src_nids()
    tgt_node_ids = model.get_tgt_nids()
    p_model = model.apply(src_nid=src_node_ids, tgt_nid=tgt_node_ids)
    log.log_assert(
        np.array_equal(np.unique(p_model), [0.0, 1.0]), "ERROR: Model output not boolean!"
    )
    res_dict = {"data": csc_matrix(p_model).astype(bool)}

    # Visualize model output
    if model.is_inverted():
        model_str = "Inverted adjacency model"
    else:
        model_str = "Adjacency model"
    plt.figure(figsize=(4, 4))
    adjacency.plot(
        res_dict,
        None,
        fig_title=f"{model_str}\n({len(src_node_ids)}x{len(tgt_node_ids)} neurons, {np.sum(p_model > 0)} connections)",
        vmin=0.0,
        vmax=1.0,
    )
    plt.tight_layout()
    out_fn = os.path.abspath(os.path.join(out_dir, "model_output.png"))
    log.info(f"Saving {out_fn}...")
    plt.savefig(out_fn)



# This file is part of connectome-manipulator.
#
# SPDX-License-Identifier: Apache-2.0
# Copyright (c) 2024 Blue Brain Project/EPFL

import numpy as np
import os
import pandas as pd
import pytest

from bluepysnap import Circuit

from utils import TEST_DATA_DIR
from connectome_manipulator.connectome_manipulation.manipulation import Manipulation
from connectome_manipulator.connectome_manipulation.converters import EdgeWriter
from connectome_manipulator import log


@pytest.fixture
def manipulation():
    m = Manipulation.get("conn_removal")
    return m


def test_apply(manipulation):
    log.setup_logging()  # To have data logging in a defined state

    circuit = Circuit(os.path.join(TEST_DATA_DIR, "circuit_sonata.json"))
    edges = circuit.edges[circuit.edges.population_names[0]]
    nodes = [edges.source, edges.target]
    src_ids = nodes[0].ids()
    tgt_ids = nodes[1].ids()
    edges_table = edges.afferent_edges(tgt_ids, properties=edges.property_names)

    # Test that given percentage of connections is removed
    total_conns = len(np.unique(edges_table[["@source_node", "@target_node"]], axis=0))
    for pct in range(0, 101, 25):
        writer = EdgeWriter(None, existing_edges=edges_table.copy())
        manipulation(nodes, writer).apply(tgt_ids, amount_pct=pct)
        res = writer.to_pandas()
        res_conns = len(np.unique(res[["@source_node", "@target_node"]], axis=0))
        assert res_conns == int(total_conns * (100 - pct) / 100)

    # Check that only given ids are considered
    src_id = 0
    writer = EdgeWriter(None, existing_edges=edges_table.copy())
    manipulation(nodes, writer).apply(tgt_ids, sel_src={*src_ids} - {src_id}, amount_pct=100)
    res = writer.to_pandas()
    assert np.all(res["@source_node"] == src_id)

    tgt_id = 9
    writer = EdgeWriter(None, existing_edges=edges_table.copy())
    manipulation(nodes, writer).apply(tgt_ids, sel_dest={*tgt_ids} - {tgt_id}, amount_pct=100)
    res = writer.to_pandas()
    assert np.all(res["@target_node"] == tgt_id)

    # Check when both sel_src and sel_dest are given
    writer = EdgeWriter(None, existing_edges=edges_table.copy())
    manipulation(nodes, writer).apply(
        tgt_ids, sel_src={*src_ids} - {src_id}, sel_dest={*tgt_ids} - {tgt_id}, amount_pct=100
    )
    res = writer.to_pandas()
    assert np.all(np.logical_or(res["@source_node"] == src_id, res["@target_node"] == tgt_id))

    # Check that connection size filtering works for minimum connection size
    writer = EdgeWriter(None, existing_edges=edges_table.copy())
    manipulation(nodes, writer).apply(tgt_ids, min_syn_per_conn=2, amount_pct=100)
    res = writer.to_pandas()
    _, n_syn_conn = np.unique(res[["@source_node", "@target_node"]], axis=0, return_counts=True)
    assert np.all(n_syn_conn < 2)

    # Check that connection size filtering works for maximum connection size
    writer = EdgeWriter(None, existing_edges=edges_table.copy())
    manipulation(nodes, writer).apply(tgt_ids, max_syn_per_conn=2, amount_pct=100)
    res = writer.to_pandas()
    _, n_syn_conn = np.unique(res[["@source_node", "@target_node"]], axis=0, return_counts=True)
    assert np.all(n_syn_conn > 2)

    # Check that connection size filtering works when both are enabled
    writer = EdgeWriter(None, existing_edges=edges_table.copy())
    manipulation(nodes, writer).apply(
        tgt_ids, min_syn_per_conn=2, max_syn_per_conn=2, amount_pct=100
    )
    res = writer.to_pandas()
    _, n_syn_conn = np.unique(res[["@source_node", "@target_node"]], axis=0, return_counts=True)
    assert np.all(n_syn_conn != 2)

    # Check with empty selection, i.e. nothing to be removed
    writer = EdgeWriter(None, existing_edges=edges_table.copy())
    manipulation(nodes, writer).apply(tgt_ids, min_syn_per_conn=np.inf, amount_pct=100)
    res = writer.to_pandas()
    _, n_syn_conn = np.unique(res[["@source_node", "@target_node"]], axis=0, return_counts=True)
    assert res.equals(edges_table)



# This file is part of connectome-manipulator.
#
# SPDX-License-Identifier: Apache-2.0
# Copyright (c) 2024 Blue Brain Project/EPFL

"""Customized profling"""

import resource
import time
import os
import logging
from contextlib import ContextDecorator
import pandas as pd


logger_profiling = logging.getLogger(__name__)


class _ResourceProfiler:
    """CPU/Memory profiling class"""

    def __init__(self, enabled=False):
        """Class initialization."""
        self._enabled = enabled

        self._start_mem = None
        self.diff_mem = None

        self._start_time = None
        self.diff_time = None

    def start(self):
        """Start profiling."""
        if not self._enabled:
            return

        self._start_mem = resource.getrusage(resource.RUSAGE_SELF).ru_maxrss / 1024**2
        self._start_time = time.perf_counter()

    def stop(self):
        """Stop profiling."""
        if not self._enabled:
            return
        self.diff_time = time.perf_counter() - self._start_time
        self._start_time = None  # invalidate start time

        self.diff_mem = (
            resource.getrusage(resource.RUSAGE_SELF).ru_maxrss / 1024**2 - self._start_mem
        )
        self._start_mem = None  # invalidate start memory


class _ProfilerManager:
    """CPU/Memory profiling manager class"""

    def __init__(self):
        self.profilers = {}
        self._enable = False
        self.perf_table = None
        self._csv_file = None
        self._parent_labels = []

        self.init_perf_table()
        logger_profiling.setLevel(logging.INFO)

    def init_perf_table(self):
        """Initialize the performance table."""
        self.perf_table = pd.DataFrame(
            [],
            columns=["label", "time", "memory", "parent_labels"],
        )
        self.perf_table.index.name = "id"

    def set_csv_file(self, csv_file=None):
        """Set the profiling output csv file"""
        if csv_file is None:
            return
        if not os.path.exists(os.path.split(csv_file)[0]):
            os.makedirs(os.path.split(csv_file)[0])
        self._csv_file = csv_file

    def start(self, name):
        """Starts profiling"""
        self._parent_labels.append(name)
        self.profilers.setdefault(name, []).append(_ResourceProfiler(enabled=self._enable))
        self.profilers[name][-1].start()

    def stop(self, name):
        """Stops profiling"""
        if name not in self.profilers:
            raise KeyError("{} not initialized in profilers dict".format(name))
        pinfo = self.profilers[name][-1]
        pinfo.stop()
        self._parent_labels.pop()
        # Update the performance table with the new row of profiling data
        self.perf_table.loc[self.perf_table.shape[0]] = [
            name,
            pinfo.diff_time,
            pinfo.diff_mem,
            ":".join(self._parent_labels),
        ]

    def show_stats(self):
        """Logs profiling stats"""
        if not self._enable:
            return

        # Simplify perf table and add min, max, avg
        self.perf_table = (
            self.perf_table.groupby("label")
            .agg(
                {
                    "time": ["min", "mean", "max"],
                    "memory": ["min", "mean", "max"],
                    "parent_labels": "first",
                }
            )
            .reset_index()
        )

        delim = "\u255a"
        stats_name = " PROFILER STATS "
        logger_profiling.info("+{:=^110s}+".format(stats_name))
        logger_profiling.info(
            "|{:^44s}|{:^10s}|{:^10s}|{:^10s}|{:^10s}|{:^10s}|{:^10s}|".format(
                "Event Label", "Min.Time", "Avg.Time", "Max.Time", "Min.Mem", "Avg.Mem", "Max.Mem"
            )
        )
        logger_profiling.info("+{:-^110s}+".format("-"))

        # Loop through the labels in order
        for label_name in self.profilers:
            # Access the row in perf_table based on the label
            row = self.perf_table.loc[self.perf_table["label"] == label_name]
            label = row["label"].item().strip()
            time_min = row[("time", "min")].item()
            time_mean = row[("time", "mean")].item()
            time_max = row[("time", "max")].item()
            mem_min = row[("memory", "min")].item()
            mem_mean = row[("memory", "mean")].item()
            mem_max = row[("memory", "max")].item()
            parents = row[("parent_labels", "first")].item()
            num_tabs = len(parents.split(":")) if parents else 0
            base_name = (
                "  " * num_tabs + delim.join("  ") * label.count(delim) + label.split(delim)[-1]
            )
            logger_profiling.info(
                "| {:<42s} | {:8.2f} | {:8.2f} | {:8.2f} | {:8.2f} | {:8.2f} | {:8.2f} |".format(
                    base_name, time_min, time_mean, time_max, mem_min, mem_mean, mem_max
                )
            )
        self.write_to_csv()
        logger_profiling.info("+{:-^110s}+".format("-"))

    def write_to_csv(self):
        """Add entry to performance table and write to .csv file."""
        if self._csv_file is None:
            return

        self.perf_table.to_csv(self._csv_file, index=False)

    def set_enabled(self, enable):
        """Enables profiling"""
        self._enable = enable

    def merge(self, profiler_manager):
        """Merges profile managers"""
        if not self._enable:
            return

        self.profilers.update(profiler_manager.profilers)
        if self.perf_table is not None and profiler_manager.perf_table is not None:
            self.perf_table = pd.concat([self.perf_table, profiler_manager.perf_table])
        elif profiler_manager.perf_table is not None:
            self.perf_table = profiler_manager.perf_table


ProfilerManager = _ProfilerManager()  # singleon


# Can be used as context manager or decorator
class profileit(ContextDecorator):
    """Context manager or decorator of for profiling"""

    def __init__(self, name):
        """Initialization"""
        self._name = name

    def __enter__(self):
        """Enter the context and starts profiling"""
        ProfilerManager.start(self._name)

    def __exit__(self, exc_type, exc, exc_tb):
        """Leave the context and stops profiling"""
        ProfilerManager.stop(self._name)



# This file is part of connectome-manipulator.
#
# SPDX-License-Identifier: Apache-2.0
# Copyright (c) 2024 Blue Brain Project/EPFL

"""Synapse removal module."""

import numpy as np

from scipy.sparse import csc_matrix

from connectome_manipulator import log, profiler
from connectome_manipulator.access_functions import get_node_ids
from connectome_manipulator.connectome_manipulation.manipulation import Manipulation


class SynapseRemoval(Manipulation):
    """Connectome manipulation class for removing synapses:

    Removes a percentage of randomly selected synapses according to certain cell
    selection criteria, optionally keeping connections (i.e., at least 1 syn/conn)
    and rescaling g_syns to keep sum of g_syns per connection constant (unless there
    is no synapse per connection left).
    The manipulation can be applied through the :func:`apply` method.
    """

    @profiler.profileit(name="syn_removal")
    def apply(
        self,
        split_ids,
        sel_src=None,
        sel_dest=None,
        amount_pct=100.0,
        keep_conns=False,
        rescale_gsyn=False,
        **kwargs,
    ):
        """Applies a removal of synapses according to certain cell selection criteria.

        Args:
            split_ids (list-like): List of neuron IDs that are part of the current data split; will be automatically provided by the manipulator framework
            sel_src (str/list-like/dict): Source (pre-synaptic) neuron selection
            sel_dest (str/list-like/dict): Target (post-synaptic) neuron selection
            amount_pct (float): Percentage of randomly sampled synapses to be removed
            keep_conns (bool): If selected, won't remove entire connections but will keep (at least) one synapse per connection
            rescale_gsyn (bool): If selected, rescales g_syn (synaptic "conductance" property) so that the sum of g_syns per connections before and after the manipulation is kept the same (unless the entire connection has been removed)
            **kwargs: Additional keyword arguments - Not used

        Note:
            Input/output edges (synapse) tables are accessed through the ``writer`` object:

            * Loading input edges: ``edges_table = self.writer.to_pandas()``
            * Writing output edges: ``self.writer.from_pandas(edges_table_manip)``
        """
        # pylint: disable=arguments-differ
        log.log_assert(0.0 <= amount_pct <= 100.0, "amount_pct out of range!")

        gids_src = get_node_ids(self.nodes[0], sel_src)
        gids_dest = get_node_ids(self.nodes[1], sel_dest)

        edges_table = self.writer.to_pandas()
        syn_sel_idx = np.logical_and(
            np.isin(edges_table["@source_node"], gids_src),
            np.isin(edges_table["@target_node"], gids_dest),
        )  # All potential synapses to be removed

        if rescale_gsyn:
            # Determine connection strength (sum of g_syns per connection) BEFORE synapse removal
            gsyn_table = self._get_gsyn_sum_per_conn(edges_table, gids_src, gids_dest)

        if keep_conns:  # Keep (at least) one synapse per connection
            rnd_perm = np.random.permutation(np.sum(syn_sel_idx))
            _, syn_idx_to_keep = np.unique(
                edges_table[syn_sel_idx].iloc[rnd_perm][["@source_node", "@target_node"]],
                axis=0,
                return_index=True,
            )  # Randomize order, so that index of first occurrence is randomized

            syn_keep_idx = np.ones(np.sum(syn_sel_idx)).astype(bool)
            syn_keep_idx[syn_idx_to_keep] = False
            inv_perm = np.argsort(rnd_perm)
            syn_sel_idx[syn_sel_idx] = syn_keep_idx[inv_perm]  # Restore original order

        num_syn = np.sum(syn_sel_idx)
        num_remove = np.round(amount_pct * num_syn / 100).astype(int)

        log.info(
            f"Removing {num_remove} ({amount_pct}%) of {num_syn} synapses (sel_src={sel_src}, sel_dest={sel_dest}, keep_conns={keep_conns}, rescale_gsyn={rescale_gsyn})"
        )

        sel_remove = np.random.permutation([True] * num_remove + [False] * (num_syn - num_remove))
        syn_sel_idx[syn_sel_idx] = sel_remove  # Set actual indices of synapses to be removed
        edges_table_manip = edges_table[~syn_sel_idx].copy()

        if rescale_gsyn:
            # Determine connection strength (sum of g_syns per connection) AFTER synapse removal ...
            gsyn_table_manip = self._get_gsyn_sum_per_conn(edges_table_manip, gids_src, gids_dest)

            # ... and rescale g_syn so that the sum of g_syns per connections BEFORE and AFTER manipulation is kept the same (unless there is no synapse per connection left)
            self._rescale_gsyn_per_conn(
                edges_table_manip, gids_src, gids_dest, gsyn_table, gsyn_table_manip
            )

        self.writer.from_pandas(edges_table_manip)

    @staticmethod
    def _get_gsyn_sum_per_conn(edges_table, gids_src, gids_dest):
        """Helper function to determine sum of g_syns per connection."""
        mask = np.logical_and(
            np.in1d(edges_table["@source_node"], gids_src),
            np.in1d(edges_table["@target_node"], gids_dest),
        )
        gsyn_tab = (
            edges_table.loc[mask].groupby(["@source_node", "@target_node"]).sum()["conductance"]
        )
        gsyn_mat = csc_matrix(
            (
                gsyn_tab.to_numpy(),
                (
                    gsyn_tab.index.get_level_values("@source_node").to_numpy(),
                    gsyn_tab.index.get_level_values("@target_node").to_numpy(),
                ),
            )
        )

        return gsyn_mat

    @staticmethod
    def _rescale_gsyn_per_conn(edges_table, gids_src, gids_dest, gsyn_mat, gsyn_mat_manip):
        """Helper function to rescale g_syn in case of changed number of synapses per connection, keeping sum of g_syns per connection constant."""
        mask = np.logical_and(
            np.isin(edges_table["@source_node"], gids_src),
            np.isin(edges_table["@target_node"], gids_dest),
        )
        edge_indices = edges_table.index[mask]

        for edge_id in edge_indices:
            sidx = edges_table.at[edge_id, "@source_node"].astype(int)
            didx = edges_table.at[edge_id, "@target_node"].astype(int)
            if (
                gsyn_mat[sidx, didx] > 0.0
                and gsyn_mat_manip[sidx, didx] > 0.0
                and gsyn_mat_manip[sidx, didx] != gsyn_mat[sidx, didx]
            ):
                scale = gsyn_mat[sidx, didx] / gsyn_mat_manip[sidx, didx]
                edges_table.at[
                    edge_id, "conductance"
                ] *= scale  # Re-scale conductance 'in-place' in edges_table



# This file is part of connectome-manipulator.
#
# SPDX-License-Identifier: Apache-2.0
# Copyright (c) 2024 Blue Brain Project/EPFL

"""Module for comparing connectomes based on adjacency matrices:

Structural comparison of two connectomes in terms of adjacency matrices for selected pathways
(including synapse counts per connection), as specified by the config. For each connectome,
the underlying adjacency/count matrices are computed by the :func:`compute` function and will be saved
to a data file first. The individual adjacency/count matrices, together with a difference map
between the two connectomes, are then plotted by means of the :func:`plot` function.
"""

import matplotlib.pyplot as plt
import numpy as np
from scipy.sparse import csc_matrix
from connectome_manipulator.access_functions import (
    get_edges_population,
    get_node_ids,
    get_connections,
)


def compute(circuit, sel_src=None, sel_dest=None, edges_popul_name=None, **_):
    """Extracts adjacency and count matrices from a given circuit's connectome.

    Args:
        circuit (bluepysnap.Circuit): Input circuit
        sel_src (str/list-like/dict): Source (pre-synaptic) neuron selection
        sel_dest (str/list-like/dict): Target (post-synaptic) neuron selection
        edges_popul_name (str): Name of SONATA egdes population to extract data from

    Returns:
        dict: Dictionary containing the extracted data elements; see Notes

    Note:
        The returned dictionary contains the following data elements that can be selected for plotting through the structural comparison configuration file, together with a common dictionary containing additional information. Each data element is a dictionary with "data" (scipy.sparse.csc_matrix), "name" (str), and "unit" (str) items.

        * "adj": Adjacency matrix
        * "adj_cnt": Synaptome matrix, containing the numbers of synapses per connection
    """
    # Select edge population
    edges = get_edges_population(circuit, edges_popul_name)

    # Select corresponding source/target nodes populations
    src_nodes = edges.source
    tgt_nodes = edges.target

    src_node_ids = get_node_ids(src_nodes, sel_src)
    tgt_node_ids = get_node_ids(tgt_nodes, sel_dest)

    assert (
        len(src_node_ids) > 0 and len(tgt_node_ids) > 0
    ), "ERROR: Empty src/tgt node selection(s)!"

    # Map source/target node ids to continuous range of indices for plotting
    src_gid_min = min(src_nodes.ids())
    src_gid_max = max(src_nodes.ids())
    tgt_gid_min = min(tgt_nodes.ids())
    tgt_gid_max = max(tgt_nodes.ids())

    src_plot_ids = np.full(src_gid_max - src_gid_min + 1, -1).astype(int)
    src_gid_offset = src_gid_min
    src_plot_ids[src_node_ids - src_gid_offset] = np.arange(len(src_node_ids))

    def src_gid_to_idx(gids):
        return src_plot_ids[gids - src_gid_offset]

    tgt_plot_ids = np.full(tgt_gid_max - tgt_gid_min + 1, -1).astype(int)
    tgt_gid_offset = tgt_gid_min
    tgt_plot_ids[tgt_node_ids - tgt_gid_offset] = np.arange(len(tgt_node_ids))

    def tgt_gid_to_idx(gids):
        return tgt_plot_ids[gids - tgt_gid_offset]

    print(
        f"INFO: Creating {len(src_node_ids)}x{len(tgt_node_ids)} adjacency matrix (sel_src={sel_src}, sel_dest={sel_dest})",
        flush=True,
    )

    conns = get_connections(edges, src_node_ids, tgt_node_ids, with_nsyn=True)
    if len(conns) == 0:  # No connections, creating empty matrix
        count_matrix = csc_matrix((len(src_node_ids), len(tgt_node_ids)), dtype=int)
    else:
        count_matrix = csc_matrix(
            (conns[:, 2], (src_gid_to_idx(conns[:, 0]), tgt_gid_to_idx(conns[:, 1]))),
            shape=(len(src_node_ids), len(tgt_node_ids)),
            dtype=int,
        )

    adj_matrix = count_matrix > 0

    return {
        "adj": {"data": adj_matrix, "name": "Adjacency", "unit": None},
        "adj_cnt": {"data": count_matrix, "name": "Adjacency count", "unit": "Synapse count"},
        "common": {"src_gids": src_node_ids, "tgt_gids": tgt_node_ids},
    }


def plot(
    res_dict, _common_dict, fig_title=None, vmin=None, vmax=None, isdiff=False, **_
):  # pragma:no cover
    """Plots an adjacency/count matrix or a difference matrix.

    Args:
        res_dict (dict): Results dictionary, containing selected data for plotting; must contain a "data" item with a sparse matrix of type scipy.sparse.csc_matrix, as well as "name" and "unit" items containing strings.
        _common_dict (dict): Common dictionary, containing additional information - Not used
        fig_title (str): Optional figure title
        vmin (float): Minimum plot range
        vmax (float): Maximum plot range
        isdiff (bool): Flag indicating that ``res_dict`` contains a difference matrix; in this case, a symmetric plot range is required and a divergent colormap will be used
    """
    if isdiff:  # Difference plot
        assert -1 * vmin == vmax, "ERROR: Symmetric plot range required!"
        cmap = "PiYG"  # Symmetric (diverging) colormap
    else:  # Regular plot
        assert vmin == 0, "ERROR: Plot range including 0 required!"
        cmap = "hot_r"  # Regular colormap [color at 0 should be white (not actually drawn), to match figure background!]

    mat = res_dict["data"].tocoo()  # Convert to COO, for easy access to row/col and data!!
    col_idx = mat.data
    plt.scatter(
        mat.col,
        mat.row,
        marker=",",
        s=0.1,
        edgecolors="none",
        alpha=0.5,
        c=col_idx,
        cmap=cmap,
        vmin=vmin,
        vmax=vmax,
    )

    if not res_dict["data"].dtype == bool:
        cb = plt.colorbar()
        cb.set_label(res_dict["unit"])

    if fig_title is None:
        plt.title(res_dict["name"])
    else:
        plt.title(fig_title)

    plt.xlabel("Postsynaptic neurons")
    plt.ylabel("Presynaptic neurons")

    plt.axis("image")
    plt.xlim((-0.5, res_dict["data"].shape[1] - 0.5))
    plt.ylim((-0.5, res_dict["data"].shape[0] - 0.5))
    plt.gca().invert_yaxis()



# This file is part of connectome-manipulator.
#
# SPDX-License-Identifier: Apache-2.0
# Copyright (c) 2024 Blue Brain Project/EPFL

"""A module implementing several executor wrappers"""

from concurrent.futures import ProcessPoolExecutor
from contextlib import contextmanager
from datetime import datetime, timedelta
from distributed import as_completed, WorkerPlugin

from .. import log


class DaskExecutor:
    """A executor wrapper for Dask"""

    _PROCESS_INTERVAL = timedelta(minutes=1)

    def __init__(self, executor, result_hook=None) -> None:
        """Initializes the executor wrapper"""
        self._executor = executor
        self._result_hook = result_hook
        self._jobs = []
        self._last_processing = datetime.now()

    def _timed_process_jobs(self) -> bool:
        if datetime.now() - self._last_processing > self._PROCESS_INTERVAL:
            self.process_jobs(self._jobs)
            # Use the time after processing to get an evenly timed job submission window
            self._last_processing = datetime.now()

    def submit(self, func, args, extra_data):
        """Submits a new routine to be run by the distributed framework"""
        job = self._executor.submit(func, *args)
        job.extra_data = extra_data
        self._timed_process_jobs()
        self._jobs.append(job)

    def process_jobs(self, jobs=None):
        """Process completed jobs, blocking to wait for all jobs if none are passed"""
        if not jobs:
            jobs = as_completed(self._jobs)
        self._jobs = []
        for job in jobs:
            if job.done():
                if self._result_hook:
                    self._result_hook(job.result(), job.extra_data)
                job.release()
            else:
                self._jobs.append(job)


class SerialExecutor:
    """The serial executor wrapper, which immediately runs the user function"""

    def __init__(self, result_hook) -> None:
        """Initializes the serial executor wrapper"""
        self._result_hook = result_hook

    def submit(self, func, args, extra_data):
        """Submits a new routine (which is run immediately)"""
        result = func(*args)  # Run it inplace
        if self._result_hook:
            self._result_hook(result, extra_data)
        return result


@contextmanager
def serial_ctx(result_hook):
    """A plain serial executor, basically no-op"""
    yield SerialExecutor(result_hook)
    # DONE!


# This allegedly provides a bit safer support for separate processes, skip for now to keep
# dependencies at bay
# from loky import ProcessPoolExecutor
class AddProcessPool(WorkerPlugin):
    """Helper to avoid threading and use processes instead"""

    def setup(self, worker):
        """Configures the worker"""
        if worker.state.nthreads > 1:
            worker.executors["default"] = ProcessPoolExecutor(max_workers=worker.state.nthreads)


@contextmanager
def dask_ctx(result_hook, executor_params: dict):
    """An executor using the Dask system"""
    from dask.distributed import Client

    # Dask requires numeric params to go as the native type
    for k, val in executor_params.items():
        if val.isdecimal():
            try:
                executor_params[k] = int(val)
            except ValueError:
                executor_params[k] = float(val)

    with Client(**executor_params) as client:
        client.register_worker_plugin(AddProcessPool())
        executor_wrapper = DaskExecutor(client, result_hook)

        yield executor_wrapper

        log.info("Jobs submitted to DASK")

        executor_wrapper.process_jobs()

        log.info("DASK jobs finished")

        log.info("Shutting down DASK workers gracefully")
        client.retire_workers()
        # In some cases, we seem to need to shut down the client: while using dask-mpi
        # though, it will attempt to shut down the client itself. If the code below is
        # activated while dask-mpi is used, exceptions will be displayed.
        # time.sleep(1)
        # client.shutdown()


def in_context(options, params, result_hook=None):
    """An auto-selector of the executor context manager"""
    log.info("Starting Execution context")
    if options.parallel:
        return dask_ctx(result_hook, params)
    return serial_ctx(result_hook)



# This file is part of connectome-manipulator.
#
# SPDX-License-Identifier: Apache-2.0
# Copyright (c) 2024 Blue Brain Project/EPFL

"""Synapse subsampling module."""

import numpy as np

from connectome_manipulator import log, profiler
from connectome_manipulator.connectome_manipulation.manipulation import Manipulation


class SynapseSubsampling(Manipulation):
    """Connectome manipulation class for synapse subsampling:

    Subsamples synapses, keeping a certain percentage of synapses.
    The manipulation can be applied through the :func:`apply` method.
    For a more fine-grained control, please use the
    :func:`connectome_manipulator.connectome_manipulation.manipulation.syn_removal.SynapseRemoval`
    operation.
    """

    @profiler.profileit(name="syn_subsampling")
    def apply(self, split_ids, keep_pct=100.0, **kwargs):
        """Applies a random subsampling of synapses, keeping a certain percentage of synapses.

        Args:
            split_ids (list-like): List of neuron IDs that are part of the current data split; will be automatically provided by the manipulator framework
            keep_pct (float): Percentage of randomly sampled synapses to be kept
            **kwargs: Additional keyword arguments - Not used

        Note:
            Input/output edges (synapse) tables are accessed through the ``writer`` object:

            * Loading input edges: ``edges_table = self.writer.to_pandas()``
            * Writing output edges: ``self.writer.from_pandas(edges_table_manip)``

        """
        # pylint: disable=arguments-differ
        log.log_assert(0.0 <= keep_pct <= 100.0, "keep_pct out of range!")

        edges_table = self.writer.to_pandas()
        num_syn = edges_table.shape[0]
        num_keep = np.round(keep_pct * num_syn / 100).astype(int)

        log.info(f"Synapse subsampling, keeping {num_keep} ({keep_pct}%) of {num_syn} synapses")

        syn_sel_idx = np.random.permutation([True] * num_keep + [False] * (num_syn - num_keep))
        edges_table_manip = edges_table[syn_sel_idx].copy()

        self.writer.from_pandas(edges_table_manip)



# This file is part of connectome-manipulator.
#
# SPDX-License-Identifier: Apache-2.0
# Copyright (c) 2024 Blue Brain Project/EPFL

"""Main CLI entry point for parallel processing"""

import logging
import os
import resource
import socket

import dask
from dask_mpi import initialize

from .cli import app, manipulate_connectome

logging.getLogger("distributed").setLevel(logging.WARNING)


def setup():
    """Dask parallel initialization

    This allows us to bail out early if we are not the root MPI rank that will start the scheduler.
    """
    # Dask likes to eat file descriptors for breakfast, lunch, diner, so give it the most we
    # can get
    _, hard = resource.getrlimit(resource.RLIMIT_NOFILE)
    resource.setrlimit(resource.RLIMIT_NOFILE, (hard, hard))

    _TMP = os.environ.get("SHMDIR", os.environ.get("TMPDIR", os.getcwd()))

    dask.config.set(
        {
            "distributed.worker.memory.target": False,
            "distributed.worker.memory.spill": False,
            "distributed.worker.memory.pause": 0.8,
            "distributed.worker.memory.terminate": 0.95,
            "distributed.worker.use-file-locking": False,
            "distributed.worker.profile.interval": "1s",
            "distributed.worker.profile.cycle": "10s",
            "distributed.admin.tick.limit": "1m",
            "temporary-directory": _TMP,
        }
    )

    # Using 1 thread avoids issues with non-thread-save RNG etc, we are modifying the Dask
    # process pool in executors.py to use a process pool, not a thread pool.
    threads = int(os.environ.get("SLURM_CPUS_PER_TASK", 1))

    dash_url = f"{socket.getfqdn()}:8787"

    is_client = initialize(
        dashboard_address=dash_url,
        worker_class="distributed.Worker",
        worker_options={
            "local_directory": _TMP,
            "nthreads": threads,
        },
        exit=True,
    )

    if is_client:
        print("Dashboard URL:", dash_url)

        # This is a bit hackish, but seems to be a good way to change the default for the
        # application: we always want to be parallel when starting our own distributed Dask
        for param in manipulate_connectome.params:
            if param.name == "parallel":
                param.default = True

        if __name__ == "__main__":
            app()  # pylint: disable=no-value-for-parameter


setup()
